"""
Views to display Metric data.

Note on Implementation: Since this app doesn't really use django's Models,
the views here are extended from ``TemplateView`` in order to keep things
simple.

"""
from __future__ import unicode_literals
from datetime import datetime

from django.contrib.auth.decorators import user_passes_test
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.generic.edit import FormView
from django.views.generic import TemplateView

from .utils import get_r
from .forms import AggregateMetricForm, MetricCategoryForm


class ProtectedTemplateView(TemplateView):
    """Ensures that Users are authenticated and that they're staff users. This
    is used as a parent class for the rest of the views in this app."""

    _logged_in_staff = lambda u: u.is_authenticated and u.is_staff

    @method_decorator(user_passes_test(_logged_in_staff))
    def dispatch(self, *args, **kwargs):
        return super(ProtectedTemplateView, self).dispatch(*args, **kwargs)


class ProtectedFormView(FormView):
    """Ensures that Users are authenticated and that they're staff users. This
    is used as a parent class for the rest of the views in this app."""

    _logged_in_staff = lambda u: u.is_authenticated and u.is_staff

    @method_decorator(user_passes_test(_logged_in_staff))
    def dispatch(self, *args, **kwargs):
        return super(ProtectedFormView, self).dispatch(*args, **kwargs)


class DefaultView(ProtectedTemplateView):
    template_name = "redis_metrics/default.html"


class GaugesView(ProtectedTemplateView):
    template_name = "redis_metrics/gauges.html"

    def get_context_data(self, **kwargs):
        """Includes the Gauge slugs and data in the context."""
        data = super(GaugesView, self).get_context_data(**kwargs)
        data.update({'gauges': get_r().gauge_slugs()})
        return data


class MetricsListView(ProtectedTemplateView):
    template_name = "redis_metrics/metrics_list.html"

    def get_context_data(self, **kwargs):
        """Includes the metrics slugs in the context."""
        data = super(MetricsListView, self).get_context_data(**kwargs)

        # Metrics organized by category, like so:
        # { <category_name>: [ <slug1>, <slug2>, ... ]}
        data.update({'metrics': get_r().metric_slugs_by_category()})
        return data


class MetricDetailView(ProtectedTemplateView):
    template_name = "redis_metrics/metric_detail.html"

    def get_context_data(self, **kwargs):
        """Includes the metrics slugs in the context."""
        data = super(MetricDetailView, self).get_context_data(**kwargs)
        data['slug'] = kwargs['slug']
        data['granularities'] = list(get_r()._granularities())
        return data


class MetricHistoryView(ProtectedTemplateView):
    template_name = "redis_metrics/metric_history.html"

    def get_context_data(self, **kwargs):
        """Includes the metrics slugs in the context."""
        data = super(MetricHistoryView, self).get_context_data(**kwargs)

        # Accept GET query params for ``since``
        since = self.request.GET.get('since', None)
        if since and len(since) == 10:  # yyyy-mm-dd
            since = datetime.strptime(since, "%Y-%m-%d")
        elif since and len(since) == 19:  # yyyy-mm-dd HH:MM:ss
            since = datetime.strptime(since, "%Y-%m-%d %H:%M:%S")

        data.update({
            'since': since,
            'slug': kwargs['slug'],
            'granularity': kwargs['granularity'],
            'granularities': list(get_r()._granularities()),
        })
        return data


class AggregateFormView(ProtectedFormView):
    """Allow viewing multiple metrics at once. Displays a form for entering
    the metrics that you want to view, and displays them all on POST."""
    template_name = "redis_metrics/aggregate.html"
    form_class = AggregateMetricForm

    def get_success_url(self):
        """Reverses the ``redis_metric_aggregate_detail`` URL using
        ``self.metric_slugs`` as an argument."""
        slugs = '+'.join(self.metric_slugs)
        url = reverse('redis_metric_aggregate_detail', args=[slugs])
        # Django 1.6 quotes reversed URLs, which changes + into %2B. We want
        # want to keep the + in the url (it's ok according to RFC 1738)
        # https://docs.djangoproject.com/en/1.6/releases/1.6/#quoting-in-reverse
        return url.replace("%2B", "+")  # keep the plusses :-/

    def form_valid(self, form):
        """Pull the metrics from the submitted form, and store them as a
        list of strings in ``self.metric_slugs``.
        """
        self.metric_slugs = [k.strip() for k in form.cleaned_data['metrics']]
        return super(AggregateFormView, self).form_valid(form)


class AggregateDetailView(ProtectedTemplateView):
    template_name = "redis_metrics/aggregate_detail.html"

    def get_context_data(self, **kwargs):
        """Includes the metrics slugs in the context."""
        r = get_r()
        category = kwargs.pop('category', None)
        data = super(AggregateDetailView, self).get_context_data(**kwargs)

        if category:
            slug_set =   r._category_slugs(category)
        else:
            slug_set = set(kwargs['slugs'].split('+'))

        data['granularities'] = list(r._granularities())
        data['slugs'] = slug_set
        data['category'] = category
        return data


class AggregateHistoryView(ProtectedTemplateView):
    template_name = "redis_metrics/aggregate_history.html"

    def get_context_data(self, **kwargs):
        """Includes the metrics slugs in the context."""
        r = get_r()
        data = super(AggregateHistoryView, self).get_context_data(**kwargs)
        slug_set = set(kwargs['slugs'].split('+'))
        granularity = kwargs.get('granularity', 'daily')

        # Accept GET query params for ``since``
        since = self.request.GET.get('since', None)
        if since and len(since) == 10:  # yyyy-mm-dd
            since = datetime.strptime(since, "%Y-%m-%d")
        elif since and len(since) == 19:  # yyyy-mm-dd HH:MM:ss
            since = datetime.strptime(since, "%Y-%m-%d %H:%M:%S")

        data.update({
            'slugs': slug_set,
            'granularity': granularity,
            'since': since,
            'granularities': list(r._granularities())
        })
        return data


class CategoryFormView(ProtectedFormView):
    """Allows the user to (re)categorize existing metrics."""
    template_name = "redis_metrics/categorize.html"
    form_class = MetricCategoryForm

    def get(self, *args, **kwargs):
        """See if this view was called with a specified category."""
        self.initial = {"category_name":  kwargs.get('category_name', None)}
        return super(CategoryFormView, self).get(*args, **kwargs)

    def get_initial(self):
        """Initialize the form with the given category."""
        return self.initial

    def get_success_url(self):
        """Show the list of metrics."""
        return reverse('redis_metrics_list')

    def form_valid(self, form):
        """Get the category name/metric slugs from the form, and update the
        category so contains the given metrics."""
        form.categorize_metrics()
        return super(CategoryFormView, self).form_valid(form)



"""
This module contains the settings for this app. We'll attempt to pull values
from django's settings, falling back to the default values.
"""
from __future__ import unicode_literals
from django.conf import settings

import warnings


class AppSettings:
    """This application-specific settings class will act just like a
    dictionary or an object. It provides access to settings values via:

        app_settings['HOST']

    or as attribute values:

        app_settings.HOST

    On access, we'll first try to look up values in Django's settings, and
    if the setting is not found there, we'll provide a default value.
    """

    # list of (setting name, default value)
    _default_settings = {
        'CONNECTION_CLASS': None,
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'PASSWORD': None,
        'SOCKET_TIMEOUT': None,
        'SOCKET_CONNECTION_POOL': None,
        'MIN_GRANULARITY': 'daily',
        'MAX_GRANULARITY': 'yearly',
        'MONDAY_FIRST_DAY_OF_WEEK': False,
        'USE_ISO_WEEK_NUMBER': False,
    }

    # A mapping of our old settings names to the new name
    _old_settings = {
        'CONNECTION_CLASS': 'REDIS_METRICS_CONNECTION_CLASS',
        'HOST': 'REDIS_METRICS_HOST',
        'PORT': 'REDIS_METRICS_PORT',
        'DB': 'REDIS_METRICS_DB',
        'PASSWORD': 'REDIS_METRICS_PASSWORD',
        'SOCKET_TIMEOUT': 'REDIS_METRICS_SOCKET_TIMEOUT',
        'SOCKET_CONNECTION_POOL': 'REDIS_METRICS_SOCKET_CONNECTION_POOL',
        'MIN_GRANULARITY': 'REDIS_METRICS_MIN_GRANULARITY',
        'MAX_GRANULARITY': 'REDIS_METRICS_MAX_GRANULARITY',
        'MONDAY_FIRST_DAY_OF_WEEK': 'REDIS_METRICS_MONDAY_FIRST_DAY_OF_WEEK',
        'USE_ISO_WEEK_NUMBER': 'USE_ISO_WEEK_NUMBER',
    }

    def __getattr__(self, name):
        """Access settings as an attribute."""
        try:
            return self[name]
        except KeyError as e:
            # Failed attribute access should raise an AttributeError
            raise AttributeError(e)

    def __getitem__(self, key):
        """Access settings as a dictionary key."""

        try:
            # First try the project settings...
            return settings.REDIS_METRICS[key]
        except (AttributeError, KeyError):
            pass

        try:
            # Then, try one of the old settings keys.
            old_key = self._old_settings[key]
            msg = "{} is deprecated, use REDIS_METRICS['{}'] instead."
            warnings.warn(msg.format(old_key, key), DeprecationWarning, stacklevel=2)
            return getattr(settings, old_key)
        except AttributeError:
            pass

        # Fall back to the app's defaults.
        return getattr(settings, key, self._default_settings[key])


app_settings = AppSettings()


# All possible granularity values.
GRANULARITIES = ['seconds', 'minutes', 'hourly', 'daily', 'weekly', 'monthly', 'yearly']



from __future__ import unicode_literals
from django.urls import include, re_path
from django.contrib import admin

admin.autodiscover()

urlpatterns = [
    re_path(r'^admin/', admin.site.urls),
    re_path(r'^metrics/', include('redis_metrics.urls')),
]



from __future__ import unicode_literals
from datetime import datetime
try:
    from unittest.mock import call, patch, Mock
except ImportError:
    from mock import call, patch, Mock

from django.test import TestCase
from django.test.utils import override_settings

from ..models import R
from .. import utils


TEST_SETTINGS = {
    'HOST': 'localhost',
    'PORT': 6379,
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
    'MIN_GRANULARITY': 'seconds',
    'MAX_GRANULARITY': 'yearly',
    'MONDAY_FIRST_DAY_OF_WEEK': False,
    'USE_ISO_WEEK_NUMBER': False,
}


@override_settings(REDIS_METRICS=TEST_SETTINGS)
class TestUtils(TestCase):
    """Tests for functions in ``redis_metrics.utils``."""

    def setUp(self):
        utils._redis_model = None

    def test_get_r(self):
        # Global `_redis_model` is None by default
        r_kwargs = {
            'decode_responses': True,
            'host': 'localhost',
            'db': 0,
            'port': 6379,
            'password': None,
            'connection_pool': None,
            'socket_timeout': None
        }
        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            inst = R()
            self.assertEqual(inst.host, 'localhost')
            self.assertEqual(inst.port, 6379)
            self.assertEqual(inst.db, 0)
            self.assertEqual(inst.password, None)
        self.assertEqual(utils._redis_model, None)
        with patch("redis_metrics.models.redis.StrictRedis") as mock_redis:
            r = utils.get_r()
            self.assertIsInstance(r, R)
            self.assertEqual(r, utils._redis_model)
            mock_redis.assert_called_once_with(**r_kwargs)

    def test_set_metric(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.set_metric("test-slug", 42)
            mock_get_r.assert_has_calls([
                call(),
                call().set_metric("test-slug", 42, category=None, expire=None, date=None),
            ])

    def test_set_metric_with_category(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.set_metric("test-slug", 42, category="Woo")
            mock_get_r.assert_has_calls([
                call(),
                call().set_metric("test-slug", 42, category="Woo", expire=None, date=None),
            ])

    def test_set_metric_with_expiration(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.set_metric("test-slug", 42, expire=300)
            mock_get_r.assert_has_calls([
                call(),
                call().set_metric("test-slug", 42, category=None, expire=300, date=None),
            ])

    def test_metric(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.metric("test-slug")
            mock_get_r.assert_has_calls([
                call(),
                call().metric("test-slug", num=1, category=None, expire=None, date=None),
            ])

    def test_metric_with_category(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.metric("test-slug", category="Woo")
            mock_get_r.assert_has_calls([
                call(),
                call().metric("test-slug", num=1, category="Woo", expire=None, date=None),
            ])

    def test_metric_with_expiration(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.metric("test-slug", expire=300)
            mock_get_r.assert_has_calls([
                call(),
                call().metric("test-slug", num=1, category=None, expire=300, date=None),
            ])

    def test_metric_with_date(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.metric("test-slug", date=datetime(2000, 1, 2))
            mock_get_r.assert_has_calls([
                call(),
                call().metric("test-slug", num=1, category=None, expire=None, date=datetime(2000, 1, 2)),
            ])

    def test_gauge(self):
        with patch("redis_metrics.utils.get_r") as mock_get_r:
            utils.gauge("test-slug", 9000)
            mock_get_r.assert_has_calls([
                call(),
                call().gauge("test-slug", 9000),
            ])

    def test_generate_test_metrics(self):
        keys = [
            'm:test-slug:s:2000-01-02-03-04-05',
            'm:test-slug:i:2000-01-02-03-04',
            'm:test-slug:h:2000-01-02-03',
            'm:test-slug:2000-01-02',
            'm:test-slug:w:2000-01',
            'm:test-slug:m:2000-01',
            'm:test-slug:y:2000',
        ]
        config = {
            '_build_keys.return_value': keys,
            '_metric_slugs_key': 'MSK',
            '_date_range.return_value': [datetime.utcnow()],
        }
        mock_r = Mock(**config)
        config = {'return_value': mock_r}
        with patch("redis_metrics.utils.get_r", **config) as mock_get_r:
            # When called with random = True
            with patch("redis_metrics.utils.random") as mock_random:
                mock_random.randint.return_value = 9999
                utils.generate_test_metrics(
                    slug="test-slug", num=1, randomize=True, increment_value=1
                )
                mock_get_r.assert_called_once_with()
                mock_r.r.sadd.assert_called_once_with('MSK', 'test-slug')
                mock_random.seed.assert_called_once_with()
                mock_random.randint.assert_has_calls([
                    call(0, 1), call(0, 1), call(0, 1), call(0, 1),
                ])
                mock_r.r.incr.assert_has_calls([
                    call('m:test-slug:2000-01-02', 9999),
                    call('m:test-slug:w:2000-01', 9999),
                    call('m:test-slug:m:2000-01', 9999),
                    call('m:test-slug:y:2000', 9999),
                ])

            mock_get_r.reset_mock()
            mock_r.reset_mock()

            # When called with random = False
            utils.generate_test_metrics(
                slug="test-slug", num=1, randomize=False, increment_value=1
            )
            mock_get_r.assert_called_once_with()
            mock_r.r.sadd.assert_called_once_with('MSK', 'test-slug')
            mock_r.r.incr.assert_has_calls([
                call('m:test-slug:2000-01-02', 0),
                call('m:test-slug:w:2000-01', 0),
                call('m:test-slug:m:2000-01', 0),
                call('m:test-slug:y:2000', 0),
            ])

    def test_generate_test_metrics_with_cap(self):
        keys = [
            'm:test-slug:s:2000-01-02-03-04-05',
            'm:test-slug:i:2000-01-02-03-04',
            'm:test-slug:h:2000-01-02-03',
            'm:test-slug:2000-01-02',
            'm:test-slug:w:2000-01',
            'm:test-slug:m:2000-01',
            'm:test-slug:y:2000',
        ]
        config = {
            '_build_keys.return_value': keys,
            '_metric_slugs_key': 'MSK',
            'r.get.return_value': 100,
            '_date_range.return_value': [datetime.utcnow()],
        }
        mock_r = Mock(**config)
        config = {'return_value': mock_r}
        with patch("redis_metrics.utils.get_r", **config) as mock_get_r:
            # When called with random = True
            with patch("redis_metrics.utils.random") as mock_random:
                mock_random.randint.return_value = 9999
                utils.generate_test_metrics(
                    slug="test-slug", num=1, randomize=True, cap=5, increment_value=1
                )
                mock_get_r.assert_called_once_with()
                mock_r.r.sadd.assert_called_once_with('MSK', 'test-slug')
                mock_random.seed.assert_called_once_with()
                mock_random.randint.assert_has_calls([
                    call(0, 1), call(0, 1), call(0, 1), call(0, 1),
                ])
                # Should be no increment due to cap
                mock_r.r.incr.assert_has_calls([
                    call('m:test-slug:2000-01-02', 0),
                    call('m:test-slug:w:2000-01', 0),
                    call('m:test-slug:m:2000-01', 0),
                    call('m:test-slug:y:2000', 0),
                ])

            mock_get_r.reset_mock()
            mock_r.reset_mock()

    def test_delete_test_metrics(self):
        d = datetime.utcnow()
        with patch('redis_metrics.utils.get_r') as mock_get_r:
            mock_r = mock_get_r.return_value
            mock_r._metric_slugs_key = 'MSK'
            mock_r._build_keys.return_value = ['keys']
            mock_r._date_range.return_value = [d]

            utils.delete_test_metrics(slug="test-metric", num=1)
            mock_r._build_keys.assert_called_once_with("test-metric", date=d)
            mock_r.r.srem.assert_called_once_with('MSK', 'test-metric')
            mock_r.r.delete.assert_called_once_with('keys')



from __future__ import unicode_literals
import os

BASE_PATH = os.path.dirname(__file__)
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

SITE_ID = 1
DEBUG = True
SECRET_KEY = "secret"
ROOT_URLCONF = 'redis_metrics.tests.urls'
STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_PATH, 'templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'redis_metrics',
]

REDIS_METRICS = {
    'HOST': 'localhost',
    'PORT': '6379',
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
}



"""
This app doesn't have any models, per se, but the following ``R`` class is a
lightweight wrapper around Redis.

"""
from __future__ import unicode_literals
import redis

from importlib import import_module
from collections import OrderedDict
from datetime import datetime, timedelta
from django.template.defaultfilters import slugify

from .settings import app_settings, GRANULARITIES
from .templatetags import redis_metrics_filters as template_tags


def dedupe(items):
    """Remove duplicates from a sequence (of hashable items) while maintaining
    order. NOTE: This only works if items in the list are hashable types.

    Taken from the Python Cookbook, 3rd ed. Such a great book!

    """
    seen = set()
    for item in items:
        if item not in seen:
            yield item
            seen.add(item)


class R(object):

    def __init__(self, **kwargs):
        """Creates a connection to Redis, and sets the key used to store a
        set of slugs for all metrics.

        Valid keyword arguments:

        * ``categories_key`` -- The key storing a set of all metric Categories
          (default is "categories")
        * ``metric_slugs_key`` -- The key storing a set of all metrics slugs
          (default is "metric-slugs")
        * ``gauge_slugs_key`` -- The key storing a set of all slugs for gauges
          (default is "gauge-slugs")
        * ``connection_class`` -- class to use to obtain a Redis connection (set in settings.REDIS_METRICS['CONNECTION_CLASS'])
        * ``host`` -- Redis host (set in settings.REDIS_METRICS['HOST'])
        * ``port`` -- Redis port (set in settings.REDIS_METRICS['PORT'])
        * ``db`` -- Redis DB (set in settings.REDIS_METRICS['DB'])
        * ``password`` -- Redis password (set in settings.REDIS_METRICS['PASSWORD'])
        * ``socket_timeout``   -- Redis password (set in
          settings.REDIS_METRICS['SOCKET_TIMEOUT'])
        * ``connection_pool`` -- Redis connection pool info. (set in
          settings.REDIS_METRICS['SOCKET_CONNECTION_POOL'])

        """
        self._categories_key = kwargs.get('categories_key', 'categories')
        self._metric_slugs_key = kwargs.get('metric_slugs_key', 'metric-slugs')
        self._gauge_slugs_key = kwargs.get('gauge_slugs_key', 'gauge-slugs')

        self.connection_class = kwargs.pop('connection_class', app_settings.CONNECTION_CLASS)

        if self.connection_class:
            package, module = self.connection_class.rsplit('.', 1)
            self.r = getattr(import_module(package), module)()
        else:
            self.host = kwargs.pop('host', app_settings.HOST)
            self.port = kwargs.pop('port', app_settings.PORT)
            self.db = kwargs.pop('db', app_settings.DB)
            self.password = kwargs.pop('password', app_settings.PASSWORD)
            self.socket_timeout = kwargs.pop(
                'socket_timeout',
                app_settings.SOCKET_TIMEOUT
            )
            self.connection_pool = kwargs.pop(
                'connection_pool',
                app_settings.SOCKET_CONNECTION_POOL
            )

            # Create the connection to Redis
            self.r = redis.StrictRedis(
                host=self.host,
                port=self.port,
                db=self.db,
                password=self.password,
                socket_timeout=self.socket_timeout,
                connection_pool=self.connection_pool,
                decode_responses=True
            )

    def _date_range(self, granularity, since, to=None):
        """Returns a generator that yields ``datetime.datetime`` objects from
        the ``since`` date until ``to`` (default: *now*).

        * ``granularity`` -- The granularity at which the generated datetime
          objects should be created: seconds, minutes, hourly, daily, weekly,
          monthly, or yearly
        * ``since`` -- a ``datetime.datetime`` object, from which we start
          generating periods of time. This can also be ``None``, and will
          default to the past 7 days if that's the case.
        * ``to`` -- a ``datetime.datetime`` object, from which we start
          generating periods of time. This can also be ``None``, and will
          default to now if that's the case.

        If ``granularity`` is one of daily, weekly, monthly, or yearly, this
        function gives objects at the daily level.

        If ``granularity`` is one of the following, the number of datetime
        objects returned is capped, otherwise this code is really slow and
        probably generates more data than we want:

            * hourly: returns at most 720 values (~30 days)
            * minutes: returns at most 480 values (8 hours)
            * second: returns at most 300 values (5 minutes)

        For example, if granularity is "seconds", we'll receive datetime
        objects that differ by 1 second each.

        """
        if since is None:
            since = datetime.utcnow() - timedelta(days=7)  # Default to 7 days

        if to is None:
            to = datetime.utcnow()
        elapsed = (to - since)

        # Figure out how many units to generate for the elapsed time.
        # I'm going to use `granularity` as a keyword parameter to timedelta,
        # so I need to change the wording for hours and anything > days.
        if granularity == "seconds":
            units = elapsed.total_seconds()
            units = 300 if units > 300 else units
        elif granularity == "minutes":
            units = elapsed.total_seconds() / 60
            units = 480 if units > 480 else units
        elif granularity == "hourly":
            granularity = "hours"
            units = elapsed.total_seconds() / 3600
            units = 720 if units > 720 else units
        else:
            granularity = "days"
            units = elapsed.days + 1

        return (to - timedelta(**{granularity: u}) for u in range(int(units)))

    def categories(self):
        """Returns a set of Categories under which metrics may have been
        organized."""
        return self.r.smembers(self._categories_key)

    def _category_key(self, category):
        return u"c:{0}".format(category)

    def _category_slugs(self, category):
        """Returns a set of the metric slugs for the given category"""
        key = self._category_key(category)
        slugs = self.r.smembers(key)
        return slugs

    def _categorize(self, slug, category):
        """Add the ``slug`` to the ``category``. We store category data as
        as set, with a key of the form::

            c:<category name>

        The data is set of metric slugs::

            "slug-a", "slug-b", ...

        """
        key = self._category_key(category)
        self.r.sadd(key, slug)

        # Store all category names in a Redis set, for easy retrieval
        self.r.sadd(self._categories_key, category)

    def _granularities(self):
        """Returns a generator of all possible granularities based on the
        MIN_GRANULARITY and MAX_GRANULARITY settings.
        """
        keep = False
        for g in GRANULARITIES:
            if g == app_settings.MIN_GRANULARITY and not keep:
                keep = True
            elif g == app_settings.MAX_GRANULARITY and keep:
                keep = False
                yield g
            if keep:
                yield g

    def _get_metric_key_pattern(self, granularity, slug, date):
        """ The Redis metric key and date formatting patterns for each key, by granularity"""
        patterns_by_granularity = {
            "seconds": {"key": "m:{0}:s:{1}", "date_format": "%Y-%m-%d-%H-%M-%S"},
            "minutes": {"key": "m:{0}:i:{1}", "date_format": "%Y-%m-%d-%H-%M"},
            "hourly": {"key": "m:{0}:h:{1}", "date_format": "%Y-%m-%d-%H"},
            "daily": {"key": "m:{0}:{1}", "date_format": "%Y-%m-%d"},
            "weekly": {
                "key": "m:{0}:w:{1}",
                "date_format": self._get_weekly_date_format(date),
            },
            "monthly": {"key": "m:{0}:m:{1}", "date_format": "%Y-%m"},
            "yearly": {"key": "m:{0}:y:{1}", "date_format": "%Y"},
        }
        pattern_for_granularity = patterns_by_granularity[granularity]
        fmt = pattern_for_granularity["date_format"]
        date_string = date.strftime(fmt)
        return pattern_for_granularity["key"].format(slug, date_string)

    def _get_weekly_date_format(self, date):
        if app_settings.USE_ISO_WEEK_NUMBER:
            # We can return instantly because ISO week start on monday
            return "{year}-{week_no}".format(year=date.isocalendar()[0], week_no=date.isocalendar()[1])
        return "%Y-%{0}".format('W' if app_settings.MONDAY_FIRST_DAY_OF_WEEK else 'U')

    def _build_key_patterns(self, slug, date):
        """Builds an OrderedDict of metric keys and patterns for the given slug
        and date."""
        # we want to keep the order, from smallest to largest granularity
        patts = OrderedDict()
        for g in self._granularities():
            patts[g] = self._get_metric_key_pattern(g, slug, date)
        return patts

    def _build_keys(self, slug, date=None, granularity='all'):
        """Builds redis keys used to store metrics.

        * ``slug`` -- a slug used for a metric, e.g. "user-signups"
        * ``date`` -- (optional) A ``datetime.datetime`` object used to
          generate the time period for the metric. If omitted, the current date
          and time (in UTC) will be used.
        * ``granularity`` -- Must be one of: "all" (default), "yearly",
        "monthly", "weekly", "daily", "hourly", "minutes", or "seconds".

        Returns a list of strings.

        """
        slug = slugify(slug)  # Ensure slugs have a consistent format
        if date is None:
            date = datetime.utcnow()
        patts = self._build_key_patterns(slug, date)
        if granularity == "all":
            return list(patts.values())
        return [patts[granularity]]

    def metric_slugs(self):
        """Return a set of metric slugs (i.e. those used to create Redis keys)
        for this app."""
        return self.r.smembers(self._metric_slugs_key)

    def metric_slugs_by_category(self):
        """Return a dictionary of metrics data indexed by category:

            {<category_name>: set(<slug1>, <slug2>, ...)}

        """
        result = OrderedDict()
        categories = sorted(self.r.smembers(self._categories_key))
        for category in categories:
            result[category] = self._category_slugs(category)

        # We also need to see the uncategorized metric slugs, so need some way
        # to check which slugs are not already stored.
        categorized_metrics = set([  # Flatten the list of metrics
            slug for sublist in result.values() for slug in sublist
        ])
        f = lambda slug: slug not in categorized_metrics
        uncategorized = list(set(filter(f, self.metric_slugs())))
        if len(uncategorized) > 0:
            result['Uncategorized'] = uncategorized
        return result

    def delete_metric(self, slug):
        """Removes all keys for the given ``slug``."""

        # To remove all keys for a slug, I need to retrieve them all from
        # the set of metric keys, This uses the redis "keys" command, which is
        # inefficient, but this shouldn't be used all that often.
        prefix = "m:{0}:*".format(slug)
        keys = self.r.keys(prefix)
        self.r.delete(*keys)  # Remove the metric data

        # Finally, remove the slug from the set
        self.r.srem(self._metric_slugs_key, slug)

    def set_metric(self, slug, value, category=None, expire=None, date=None):
        """Assigns a specific value to the *current* metric. You can use this
        to start a metric at a value greater than 0 or to reset a metric.

        The given slug will be used to generate Redis keys at the following
        granularities: Seconds, Minutes, Hours, Day, Week, Month, and Year.

        Parameters:

        * ``slug`` -- a unique value to identify the metric; used in
          construction of redis keys (see below).
        * ``value`` -- The value of the metric.
        * ``category`` -- (optional) Assign the metric to a Category (a string)
        * ``expire`` -- (optional) Specify the number of seconds in which the
          metric will expire.
        * ``date`` -- (optional) Specify the timestamp for the metric; default
          used to build the keys will be the current date and time in UTC form.

        Redis keys for each metric (slug) take the form:

            m:<slug>:s:<yyyy-mm-dd-hh-mm-ss> # Second
            m:<slug>:i:<yyyy-mm-dd-hh-mm>    # Minute
            m:<slug>:h:<yyyy-mm-dd-hh>       # Hour
            m:<slug>:<yyyy-mm-dd>            # Day
            m:<slug>:w:<yyyy-num>            # Week (year - week number)
            m:<slug>:m:<yyyy-mm>             # Month
            m:<slug>:y:<yyyy>                # Year

        """
        keys = self._build_keys(slug, date=date)

        # Add the slug to the set of metric slugs
        self.r.sadd(self._metric_slugs_key, slug)

        # Construct a dictionary of key/values for use with mset
        data = {}
        for k in keys:
            data[k] = value
        self.r.mset(data)

        # Add the category if applicable.
        if category:
            self._categorize(slug, category)

        # Expire the Metric in ``expire`` seconds if applicable.
        if expire:
            for k in keys:
                self.r.expire(k, expire)

    def metric(self, slug, num=1, category=None, expire=None, date=None):
        """Records a metric, creating it if it doesn't exist or incrementing it
        if it does. All metrics are prefixed with 'm', and automatically
        aggregate for Seconds, Minutes, Hours, Day, Week, Month, and Year.

        Parameters:

        * ``slug`` -- a unique value to identify the metric; used in
          construction of redis keys (see below).
        * ``num`` -- Set or Increment the metric by this number; default is 1.
        * ``category`` -- (optional) Assign the metric to a Category (a string)
        * ``expire`` -- (optional) Specify the number of seconds in which the
          metric will expire.
        * ``date`` -- (optional) Specify the timestamp for the metric; default
          used to build the keys will be the current date and time in UTC form.

        Redis keys for each metric (slug) take the form:

            m:<slug>:s:<yyyy-mm-dd-hh-mm-ss> # Second
            m:<slug>:i:<yyyy-mm-dd-hh-mm>    # Minute
            m:<slug>:h:<yyyy-mm-dd-hh>       # Hour
            m:<slug>:<yyyy-mm-dd>            # Day
            m:<slug>:w:<yyyy-num>            # Week (year - week number)
            m:<slug>:m:<yyyy-mm>             # Month
            m:<slug>:y:<yyyy>                # Year

        """
        # Add the slug to the set of metric slugs
        self.r.sadd(self._metric_slugs_key, slug)

        if category:
            self._categorize(slug, category)

        # Increment keys. NOTE: current redis-py (2.7.2) doesn't include an
        # incrby method; .incr accepts a second ``amount`` parameter.
        keys = self._build_keys(slug, date=date)

        # Use a pipeline to speed up incrementing multiple keys
        pipe = self.r.pipeline()
        for key in keys:
            pipe.incr(key, num)
            if expire:
                pipe.expire(key, expire)
        pipe.execute()

    def get_metric(self, slug):
        """Get the current values for a metric.

        Returns a dictionary with metric values accumulated for the seconds,
        minutes, hours, day, week, month, and year.

        """
        results = OrderedDict()
        granularities = self._granularities()
        keys = self._build_keys(slug)
        for granularity, key in zip(granularities, keys):
            results[granularity] = self.r.get(key)
        return results

    def get_metrics(self, slug_list):
        """Get the metrics for multiple slugs.

        Returns a list of two-tuples containing the metric slug and a
        dictionary like the one returned by ``get_metric``::

            (
                some-metric, {
                    'seconds': 0, 'minutes': 0, 'hours': 0,
                    'day': 0, 'week': 0, 'month': 0, 'year': 0
                }
            )

        """
        # meh. I should have been consistent here, but I'm lazy, so support these
        # value names instead of granularity names, but respect the min/max
        # granularity settings.
        keys = ['seconds', 'minutes', 'hours', 'day', 'week', 'month', 'year']
        key_mapping = {gran: key for gran, key in zip(GRANULARITIES, keys)}
        keys = [key_mapping[gran] for gran in self._granularities()]

        results = []
        for slug in slug_list:
            metrics = self.r.mget(*self._build_keys(slug))
            if any(metrics):  # Only if we have data.
                results.append((slug, dict(zip(keys, metrics))))
        return results

    def get_category_metrics(self, category):
        """Get metrics belonging to the given category"""
        slug_list = self._category_slugs(category)
        return self.get_metrics(slug_list)

    def delete_category(self, category):
        """Removes the category from Redis. This doesn't touch the metrics;
        they simply become uncategorized."""
        # Remove mapping of metrics-to-category
        category_key = self._category_key(category)
        self.r.delete(category_key)

        # Remove category from Set
        self.r.srem(self._categories_key, category)

    def reset_category(self, category, metric_slugs):
        """Resets (or creates) a category containing a list of metrics.

        * ``category`` -- A category name
        * ``metric_slugs`` -- a list of all metrics that are members of the
            category.

        """
        key = self._category_key(category)
        if len(metric_slugs) == 0:
            # If there are no metrics, just remove the category
            self.delete_category(category)
        else:
            # Save all the slugs in the category, and save the category name
            self.r.sadd(key, *metric_slugs)
            self.r.sadd(self._categories_key, category)

    def get_metric_history(self, slugs, since=None, to=None, granularity='daily'):
        """Get history for one or more metrics.

        * ``slugs`` -- a slug OR a list of slugs
        * ``since`` -- the date from which we start pulling metrics
        * ``to`` -- the date until which we start pulling metrics
        * ``granularity`` -- seconds, minutes, hourly,
                             daily, weekly, monthly, yearly

        Returns a list of tuples containing the Redis key and the associated
        metric::

            r = R()
            r.get_metric_history('test', granularity='weekly')
            [
                ('m:test:w:2012-52', '15'),
            ]

        To get history for multiple metrics, just provide a list of slugs::

            metrics = ['test', 'other']
            r.get_metric_history(metrics, granularity='weekly')
            [
                ('m:test:w:2012-52', '15'),
                ('m:other:w:2012-52', '42'),
            ]

        """
        if not type(slugs) == list:
            slugs = [slugs]

        # Build the set of Redis keys that we need to get.
        keys = []
        for slug in slugs:
            for date in self._date_range(granularity, since, to):
                keys += self._build_keys(slug, date, granularity)
        keys = list(dedupe(keys))

        # Fetch our data, replacing any None-values with zeros
        results = [0 if v is None else v for v in self.r.mget(keys)]
        results = zip(keys, results)
        return sorted(results, key=lambda t: t[0])

    def get_metric_history_as_columns(self, slugs, since=None,
                                      granularity='daily'):
        """Provides the same data as ``get_metric_history``, but in a columnar
        format. If you had the following yearly history, for example::

            [
                ('m:bar:y:2012', '1'),
                ('m:bar:y:2013', '2'),
                ('m:foo:y:2012', '3'),
                ('m:foo:y:2013', '4')
            ]

        this method would provide you with the following data structure::

            [
                ['Period',  'bar',  'foo']
                ['y:2012',  '1',    '3'],
                ['y:2013',  '2',    '4'],
            ]

        Note that this also includes a header column. Data in this format may
        be useful for certain graphing libraries (I'm looking at you Google
        Charts LineChart).

        """
        history = self.get_metric_history(slugs, since, granularity=granularity)
        _history = []  # new, columnar history
        periods = ['Period']  # A separate, single column for the time period
        for s in slugs:
            column = [s]  # story all the data for a single slug
            for key, value in history:
                # ``metric_slug`` extracts the slug from the Redis Key
                if template_tags.metric_slug(key) == s:
                    column.append(value)

                # Get time period value as first column; This value is
                # duplicated in the Redis key for each value, so this is a bit
                # inefficient, but... oh well.
                period = template_tags.strip_metric_prefix(key)
                if period not in periods:
                    periods.append(period)

            _history.append(column)  # Remember that slug's column of data

        # Finally, stick the time periods in the first column.
        _history.insert(0, periods)
        return list(zip(*_history))  # Transpose the rows & columns

    def get_metric_history_chart_data(self, slugs, since=None, granularity='daily'):
        """Provides the same data as ``get_metric_history``, but with metrics
        data arranged in a format that's easy to plot with Chart.js. If you had
        the following yearly history, for example::

            [
                ('m:bar:y:2012', '1'),
                ('m:bar:y:2013', '2'),
                ('m:bar:y:2014', '3'),
                ('m:foo:y:2012', '4'),
                ('m:foo:y:2013', '5')
                ('m:foo:y:2014', '6')
            ]

        this method would provide you with the following data structure::

            'periods': ['y:2012', 'y:2013', 'y:2014']
            'data': [
              {
                'slug': 'bar',
                'values': [1, 2, 3]
              },
              {
                'slug': 'foo',
                'values': [4, 5, 6]
              },
            ]

        """
        slugs = sorted(slugs)
        history = self.get_metric_history(slugs, since, granularity=granularity)

        # Convert the history into an intermediate data structure organized
        # by periods. Since the history is sorted by key (which includes both
        # the slug and the date, the values should be ordered correctly.
        periods = []
        data = OrderedDict()
        for k, v in history:
            period = template_tags.strip_metric_prefix(k)
            if period not in periods:
                periods.append(period)

            slug = template_tags.metric_slug(k)
            if slug not in data:
                data[slug] = []
            data[slug].append(v)

        # Now, reorganize data for our end result.
        metrics = {'periods': periods, 'data': []}
        for slug, values in data.items():
            metrics['data'].append({
                'slug': slug,
                'values': values
            })

        return metrics  # templates still don't like defaultdict's

    # Gauges. Gauges have a different prefix "g:" in order to differentiate
    # them from a metric of the same name.
    def gauge_slugs(self):
        """Return a set of Gauges slugs (i.e. those used to create Redis keys)
        for this app."""
        return self.r.smembers(self._gauge_slugs_key)

    def _gauge_key(self, slug):
        """Make sure our slugs have a consistent format."""
        return "g:{0}".format(slugify(slug))

    def gauge(self, slug, current_value):
        """Set the value for a Gauge.

        * ``slug`` -- the unique identifier (or key) for the Gauge
        * ``current_value`` -- the value that the gauge should display

        """
        k = self._gauge_key(slug)
        self.r.sadd(self._gauge_slugs_key, slug)  # keep track of all Gauges
        self.r.set(k, current_value)

    def get_gauge(self, slug):
        k = self._gauge_key(slug)
        return self.r.get(k)

    def delete_gauge(self, slug):
        """Removes all gauges with the given ``slug``."""
        key = self._gauge_key(slug)
        self.r.delete(key)  # Remove the Gauge
        self.r.srem(self._gauge_slugs_key, slug)  # Remove from the set of keys



from __future__ import unicode_literals
from django.test import TestCase
from django.test.utils import override_settings

from .. import settings as rs


TEST_SETTINGS = {
    'HOST': 'localhost',
    'PORT': 6379,
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
    'MIN_GRANULARITY': 'seconds',
    'MAX_GRANULARITY': 'yearly',
    'MONDAY_FIRST_DAY_OF_WEEK': False,
    'USE_ISO_WEEK_NUMBER': False,
}


@override_settings(REDIS_METRICS=TEST_SETTINGS)
class TestAppSettings(TestCase):

    def test_settings_getattr(self):
        self.assertEquals(rs.app_settings.HOST, 'localhost')
        self.assertEquals(rs.app_settings.PORT, 6379)
        self.assertEquals(rs.app_settings.DB, 0)
        self.assertEquals(rs.app_settings.PASSWORD, None)
        self.assertEquals(rs.app_settings.SOCKET_TIMEOUT, None)
        self.assertEquals(rs.app_settings.SOCKET_CONNECTION_POOL, None)
        self.assertEquals(rs.app_settings.MIN_GRANULARITY, 'seconds')
        self.assertEquals(rs.app_settings.MAX_GRANULARITY, 'yearly')
        self.assertEquals(rs.app_settings.MONDAY_FIRST_DAY_OF_WEEK, False)
        self.assertEqual(rs.app_settings.USE_ISO_WEEK_NUMBER, False)

    def test_settings_getitem(self):
        self.assertEquals(rs.app_settings['HOST'], 'localhost')
        self.assertEquals(rs.app_settings['PORT'], 6379)
        self.assertEquals(rs.app_settings['DB'], 0)
        self.assertEquals(rs.app_settings['PASSWORD'], None)
        self.assertEquals(rs.app_settings['SOCKET_TIMEOUT'], None)
        self.assertEquals(rs.app_settings['SOCKET_CONNECTION_POOL'], None)
        self.assertEquals(rs.app_settings['MIN_GRANULARITY'], 'seconds')
        self.assertEquals(rs.app_settings['MAX_GRANULARITY'], 'yearly')
        self.assertEquals(rs.app_settings['MONDAY_FIRST_DAY_OF_WEEK'], False)
        self.assertEqual(rs.app_settings['USE_ISO_WEEK_NUMBER'], False)

    def test_settings_getattr_raises_attribute_error(self):
        with self.assertRaises(AttributeError):
            rs.app_settings.LOLWUT

    def test__test_granularities_constant(self):
        # Smoke test to catch myself when this changes.
        self.assertEqual(
            rs.GRANULARITIES,
            ['seconds', 'minutes', 'hourly', 'daily', 'weekly', 'monthly', 'yearly']
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from datetime import datetime, timedelta

from django.test import TestCase
from django.test.utils import override_settings
try:
    from unittest.mock import patch
except ImportError:
    from mock import call, patch  # NOQA
from redis_metrics.templatetags import redis_metric_tags as taglib
from redis_metrics.templatetags.redis_metrics_filters import (
    metric_slug, strip_metric_prefix, to_int, to_int_list
)


TEST_SETTINGS = {
    'HOST': 'localhost',
    'PORT': 6379,
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
    'MIN_GRANULARITY': 'seconds',
    'MAX_GRANULARITY': 'yearly',
    'MONDAY_FIRST_DAY_OF_WEEK': False,
    'USE_ISO_WEEK_NUMBER': False,
}


@override_settings(REDIS_METRICS=TEST_SETTINGS)
class TestTemplateTags(TestCase):
    """Verify that template tags return the expected results."""
    maxDiff = None

    def setUp(self):
        """Patch the ``R`` class."""
        self.r_patcher = patch('redis_metrics.models.R')
        self.mock_r = self.r_patcher.start()

    def tearDown(self):
        self.r_patcher.stop()

    def test_metrics_since_history(self):
        """Tests the ``metrics_since`` template tag when displaying metric
        history."""

        slug = "test-slug"
        years = 5
        link_type = "history"
        now = datetime(2014, 7, 4)

        module = 'redis_metrics.templatetags.redis_metric_tags.datetime'
        with patch(module) as mock_datetime:
            mock_datetime.utcnow.return_value = now

            result = taglib.metrics_since(slug, years, link_type)
            self.assertIn('link_type', result.keys())
            self.assertIn('slug_values', result.keys())
            self.assertEqual(result['link_type'], link_type)

            # Verify contents of `slug_values`
            # There should be entries for each year + 5 additional periods.
            expected = [
                (slug, now - timedelta(days=1), "Today", 'daily'),
                (slug, now - timedelta(days=7), "1 Week", 'daily'),
                (slug, now - timedelta(days=30), "30 Days", 'daily'),
                (slug, now - timedelta(days=60), "60 Days", 'daily'),
                (slug, now - timedelta(days=90), "90 Days", 'daily'),
                (slug, now - timedelta(days=365), "1 Years", 'daily'),
                (slug, now - timedelta(days=365 * 2), "2 Years", 'daily'),
                (slug, now - timedelta(days=365 * 3), "3 Years", 'daily'),
                (slug, now - timedelta(days=365 * 4), "4 Years", 'daily'),
                (slug, now - timedelta(days=365 * 5), "5 Years", 'daily'),
            ]
            self.assertEqual(expected, result['slug_values'])

    def test_metrics_since_aggregate(self):
        """Tests the ``metrics_since`` template tag when displaying metric
        aggregate history."""

        slugs = ['test-a', 'test-b']
        years = 5
        link_type = "aggregate"
        granularity = "weekly"
        now = datetime(2014, 7, 4)

        module = 'redis_metrics.templatetags.redis_metric_tags.datetime'
        with patch(module) as mock_datetime:
            mock_datetime.utcnow.return_value = now

            result = taglib.metrics_since(slugs, years, link_type, granularity)
            self.assertIn('link_type', result.keys())
            self.assertIn('slug_values', result.keys())
            self.assertEqual(result['link_type'], link_type)

            # Verify contents of `slug_values`
            # There should be entries for each year + 5 additional periods.
            slugs = "+".join(slugs)
            expected = [
                (slugs, now - timedelta(days=1), "Today", granularity),
                (slugs, now - timedelta(days=7), "1 Week", granularity),
                (slugs, now - timedelta(days=30), "30 Days", granularity),
                (slugs, now - timedelta(days=60), "60 Days", granularity),
                (slugs, now - timedelta(days=90), "90 Days", granularity),
                (slugs, now - timedelta(days=365), "1 Years", granularity),
                (slugs, now - timedelta(days=365 * 2), "2 Years", granularity),
                (slugs, now - timedelta(days=365 * 3), "3 Years", granularity),
                (slugs, now - timedelta(days=365 * 4), "4 Years", granularity),
                (slugs, now - timedelta(days=365 * 5), "5 Years", granularity),
            ]
            self.assertEqual(expected, result['slug_values'])

    def test_gauge(self):
        """Tests the result of the gauge template tag."""
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            inst = mock_r.return_value
            inst.get_gauge.return_value = 100

            size = 50
            maximum = 200
            result = taglib.gauge("test-slug", maximum, size)
            expected_result = {
                'slug': "test-slug",
                'current_value': 100,
                'max_value': maximum,
                'size': size,
                'diff': maximum - 100
            }
            self.assertEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.get_gauge.assert_called_once_with("test-slug")

    def test_gauge_when_overloaded(self):
        """Tests the result of a gauge whose current value > the maximum"""
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            inst = mock_r.return_value
            inst.get_gauge.return_value = 500

            size = 50
            maximum = 200
            result = taglib.gauge("test-slug", maximum, size)
            expected_result = {
                'slug': "test-slug",
                'current_value': 500,
                'max_value': maximum,
                'size': size,
                'diff': 0,  # deff should default to 0 when overloaded.
            }
            self.assertEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.get_gauge.assert_called_once_with("test-slug")

    def test_metric_list(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            inst = mock_r.return_value
            inst.metric_slugs_by_category.return_value = "RESULT"

            result = taglib.metric_list()
            expected_result = {
                'metrics': "RESULT",
            }
            self.assertEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.metric_slugs_by_category.assert_called_once_with()

    def test_metric_detail(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            inst = mock_r.return_value
            inst._granularities.return_value = ['daily', 'weekly']
            inst.get_metric.return_value = {
                'daily': 1,
                'weekly': 2,
            }

            result = taglib.metric_detail('test')
            expected_result = {
                'granularities': ['Daily', 'Weekly'],
                'metrics': [('daily', 1), ('weekly', 2)],
                'slug': 'test',
                'with_data_table': False,
            }

            self.assertDictEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.get_metric.assert_called_once_with('test')

    def test_metric_history(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            history = [("m:test:2000-01-01", 42)]
            inst = mock_r.return_value
            inst.get_metric_history.return_value = history

            result = taglib.metric_history('test')
            expected_result = {
                'slug': 'test',
                'granularity': "daily",
                'metric_history': history,
                'since': None,
                'to': None,
                'with_data_table': False,
            }
            self.assertEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.get_metric_history.assert_called_once_with(
                slugs='test',
                granularity='daily',
                since=None,
                to=None
            )

    def test_metric_history_with_date(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            history = [("m:test:2000-01-01", 42)]
            inst = mock_r.return_value
            inst.get_metric_history.return_value = history
            expected_result = {
                'slug': 'test',
                'granularity': "daily",
                'metric_history': history,
                'since': None,
                'to': None,
                'with_data_table': False,
            }

            # With a date string
            since = datetime(2000, 1, 2)
            expected_result['since'] = since
            result = taglib.metric_history('test', since="2000-01-02")
            self.assertEqual(result, expected_result)
            inst.get_metric_history.assert_called_once_with(
                slugs='test',
                granularity='daily',
                since=since,
                to=None,
            )
            inst.reset_mock()

            # With a datetime string
            since = datetime(2000, 1, 2, 11, 30, 45)
            expected_result['since'] = since
            result = taglib.metric_history('test', since="2000-01-02 11:30:45")
            self.assertEqual(result, expected_result)
            inst.get_metric_history.assert_called_once_with(
                slugs='test',
                granularity='daily',
                since=since,
                to=None
            )
            inst.reset_mock()

            # With a datetime object
            since = datetime(2000, 1, 2, 11, 30, 45)
            expected_result['since'] = since
            result = taglib.metric_history('test', since=since)
            self.assertEqual(result, expected_result)
            inst.get_metric_history.assert_called_once_with(
                slugs='test',
                granularity='daily',
                since=since,
                to=None
            )
            inst.reset_mock()

            # With both a since and to parameter
            since = datetime(2000, 1, 2, 11, 30, 45)
            to = datetime(2001, 1, 2, 11, 30, 45)
            expected_result['since'] = since
            expected_result['to'] = to
            result = taglib.metric_history('test', since=since, to=to)
            self.assertEqual(result, expected_result)
            inst.get_metric_history.assert_called_once_with(
                slugs='test',
                granularity='daily',
                since=since,
                to=to
            )

    def test_aggregate_detail(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            slugs = ['a1', 'a2']
            inst = mock_r.return_value
            inst._granularities.return_value = ['daily', 'weekly', 'monthly', 'yearly']
            inst.get_metrics.return_value = [
                ('a1', {'day': 1, 'week': 2, 'month': 3, 'year': 4}),
                ('a2', {'day': 3, 'week': 6, 'month': 9, 'year': 12}),
            ]

            result = taglib.aggregate_detail(slugs)
            expected_result = {
                'chart_id': 'metric-aggregate-a1-a2',
                'granularities': ['Day', 'Week', 'Month', 'Year'],
                'slugs': slugs,
                'metrics': [
                    ('a1', [1, 2, 3, 4]),
                    ('a2', [3, 6, 9, 12]),
                ],
                'with_data_table': False,
            }

            self.assertDictEqual(result, expected_result)
            mock_r.assert_called_once_with()
            inst.get_metrics.assert_called_once_with(slugs)

    def test_aggregate_history(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            history = {
                'periods': ['2000-01-01', '2000-01-02', '2000-01-03'],
                'data': [
                    {
                        'slug': 'bar',
                        'values': [1, 2, 3]
                    },
                    {
                        'slug': 'foo',
                        'values': [4, 5, 6]
                    },
                ]
            }
            inst = mock_r.return_value
            inst.get_metric_history_chart_data.return_value = history

            result = taglib.aggregate_history(['foo', 'bar'])
            expected_result = {
                'chart_id': 'metric-aggregate-history-foo-bar',
                'slugs': ['foo', 'bar'],
                'since': None,
                'granularity': "daily",
                'metric_history': history,
                'with_data_table': False,
            }

            self.assertEqual(result, expected_result)
            mock_r.assert_called_once_with()  # Create the R object
            inst.get_metric_history_chart_data.assert_called_once_with(
                slugs=['foo', 'bar'],
                since=None,
                granularity='daily'
            )

    def test_aggregate_history_with_date(self):
        with patch("redis_metrics.templatetags.redis_metric_tags.get_r") as mock_r:
            history = {
                'periods': ['2000-01-01', '2000-01-02', '2000-01-03'],
                'data': [
                    {
                        'slug': 'bar',
                        'values': [1, 2, 3]
                    },
                    {
                        'slug': 'foo',
                        'values': [4, 5, 6]
                    },
                ]
            }
            inst = mock_r.return_value
            inst.get_metric_history_chart_data.return_value = history
            slugs = ['foo', 'bar']
            expected_result = {
                'chart_id': 'metric-aggregate-history-foo-bar',
                'slugs': slugs,
                'since': None,
                'granularity': "daily",
                'metric_history': history,
                'with_data_table': False,
            }

            # When given a date string
            result = taglib.aggregate_history(slugs, since="2000-01-02")
            expected_result['since'] = datetime(2000, 1, 2)
            self.assertEqual(result, expected_result)

            # When given a datetime
            result = taglib.aggregate_history(slugs, since="2000-01-02 11:30:45")
            expected_result['since'] = datetime(2000, 1, 2, 11, 30, 45)
            self.assertEqual(result, expected_result)

            # When given a datetime object
            d = datetime(2000, 1, 2, 11, 30, 45)
            result = taglib.aggregate_history(slugs, since=d)
            expected_result['since'] = d
            self.assertEqual(result, expected_result)


class TestTemplateFilters(TestCase):
    """Verify that the custom filters return expected results."""

    def test_to_int_list(self):
        """Verify that the ``int_list`` filter converts strings to numbers and that
        it returns 0 when given unexpected values."""
        self.assertEqual(to_int_list([u"3", None, "asdf", u"42"]), [3, 0, 0, 42])

    def test_to_int(self):
        """Verify that the ``int`` filter converts strings to numbers and that
        it returns 0 when given unexpected values."""
        self.assertEqual(to_int(u"3"), 3)
        self.assertEqual(to_int(u"asdf"), 0)
        self.assertEqual(to_int(None), 0)

    def test_strip_metric_prefix(self):
        # Seconds -- from: ``m:<slug>:s:<yyyy-mm-dd-hh-MM-SS>`` to ``<yyyy-mm-dd-hh-MM-SS>``
        self.assertEqual(
            strip_metric_prefix("m:test:s:2000-01-30-14-45-37"), "s:2000-01-30-14-45-37"
        )

        # Minutes -- from: ``m:<slug>:i:<yyyy-mm-dd-hh-MM>`` to ``<yyyy-mm-dd-hh-MM>``
        self.assertEqual(
            strip_metric_prefix("m:test:i:2000-01-30-14-45"), "i:2000-01-30-14-45"
        )

        # Hourly -- from: ``m:<slug>:h:<yyyy-mm-dd-hh>`` to ``<yyyy-mm-dd-hh>``
        self.assertEqual(
            strip_metric_prefix("m:test:h:2000-01-30-00"), "h:2000-01-30-00"
        )

        # Daily -- from: ``m:<slug>:<yyyy-mm-dd>`` to ``<yyyy-mm-dd>``
        self.assertEqual(
            strip_metric_prefix("m:test:2000-01-30"), "2000-01-30"
        )

        # Weekly -- from ``m:<slug>:w:<yyyy-num>`` to ``w:<yyyy-num>``
        self.assertEqual(
            strip_metric_prefix("m:test:w:2000-52"), "w:2000-52"
        )

        # Monthly -- from ``m:<slug>:m:<yyyy-mm>`` to ``m:<yyyy-mm>``
        self.assertEqual(
            strip_metric_prefix("m:test:m:2000-01"), "m:2000-01"
        )

        # Yearly -- from ``m:<slug>:y:<yyyy>`` to ``y:<yyyy>``
        self.assertEqual(
            strip_metric_prefix("m:test:y:2000"), "y:2000"
        )

    def test_metric_slug(self):
        # Converts ``m:foo:s:<yyyy-mm-dd-hh-mm-ss>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:s:2000-01-31-14-45-37"), "foo")

        # Converts ``m:foo:i:<yyyy-mm-dd-hh-mm>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:i:2000-01-31-14-45"), "foo")

        # Converts ``m:foo:h:<yyyy-mm-dd-hh>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:h:2000-01-31-14"), "foo")

        # Converts ``m:foo:<yyyy-mm-dd>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:2000-01-31"), "foo")

        # Converts ``m:foo:w:<yyyy-num>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:w:2000-52"), "foo")

        # Converts ``m:foo:m:<yyyy-mm>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:m:2000-01"), "foo")

        # Converts ``m:foo:y:<yyyy>`` to ``foo``
        self.assertEqual(metric_slug("m:foo:y:2000"), "foo")



from .test_forms import TestAggregateMetricForm, TestMetricCategoryForm
from .test_models import TestR
from .test_settings import TestAppSettings
from .test_templatetags import TestTemplateTags, TestTemplateFilters
from .test_views import TestViews
from .test_utils import TestUtils



VERSION = "2.1.0"
__version__ = VERSION

try:
    from .utils import gauge, metric, set_metric  # NOQA
except ImportError:  # pragma: no cover
    pass  # pragma: no cover

default_app_config = 'redis_metrics.apps.RedisMetricsConfig'



from __future__ import unicode_literals
from django.test import TestCase
try:
    from unittest.mock import call, patch
except ImportError:
    from mock import call, patch

from ..forms import AggregateMetricForm, MetricCategoryForm


class TestAggregateMetricForm(TestCase):

    def test_form(self):
        """Test that form has choices populated from R.metric_slugs"""
        # Set up a mock result for R.metric_slugs
        config = {'return_value.metric_slugs.return_value': ['test-slug']}
        with patch('redis_metrics.forms.R', **config) as mock_R:
            form = AggregateMetricForm()
            mock_R.assert_has_calls([
                call(),
                call().metric_slugs(),
            ])
            self.assertEqual(
                form.fields['metrics'].choices,
                [('test-slug', 'test-slug')]
            )

    def test_cleaned_data(self):
        """Verify we get expected results from cleaned_data"""
        # Set up a mock result for R.metric_slugs
        config = {'return_value.metric_slugs.return_value': ['test-slug']}
        with patch('redis_metrics.forms.R', **config):
            form = AggregateMetricForm({"metrics": ["test-slug"]})
            self.assertTrue(form.is_valid())
            self.assertEqual(form.cleaned_data, {"metrics": ["test-slug"]})


class TestMetricCategoryForm(TestCase):

    def test_form(self):
        """Test that the form has choices from R.metric_slugs, and that
        providing a ``category`` argument sets initial values."""

        # Set up a mock result for R.metric_slugs & R._category_slugs
        config = {
            'return_value.metric_slugs.return_value': ['test-slug'],
            'return_value._category_slugs.return_value': ['test-slug']
        }
        with patch('redis_metrics.forms.R', **config) as mock_R:
            # No Category
            form = MetricCategoryForm()
            self.assertFalse(form.fields['metrics'].required)

            mock_R.assert_has_calls([
                call(),
                call().metric_slugs(),
            ])
            self.assertEqual(
                form.fields['metrics'].choices,
                [('test-slug', 'test-slug')]
            )
            self.assertEqual(form.fields['metrics'].initial, None)
            self.assertEqual(form.fields['category_name'].initial, None)
            self.assertFalse(mock_R._category_slugs.called)
            mock_R.reset_mock()

            # With a Category
            initial = {'category_name': "Sample Category"}
            form = MetricCategoryForm(initial=initial)
            self.assertFalse(form.fields['metrics'].required)

            self.assertEqual(form.fields['metrics'].initial, ['test-slug'])
            self.assertEqual(
                form.fields['category_name'].initial,
                "Sample Category"
            )
            r = mock_R.return_value
            r._category_slugs.assert_called_once_with("Sample Category")

    def test_cleaned_data(self):
        """Verify we get expected results from cleaned_data."""

        # Set up a mock result for R.metric_slugs & R._category_slugs
        config = {
            'return_value.metric_slugs.return_value': ['test-slug'],
            'return_value._category_slugs.return_value': ['test-slug']
        }
        with patch('redis_metrics.forms.R', **config):
            data = {
                'category_name': 'Sample Data',
                'metrics': ['test-slug'],
            }
            form = MetricCategoryForm(data)
            self.assertTrue(form.is_valid())
            self.assertEqual(form.cleaned_data, data)

    def test_categorize_metrics(self):
        """Test the ``categorize_metrics`` method; This method should be called
        after POSTing."""
        k = {
            'return_value.metric_slugs.return_value': ['foo', 'bar', 'baz'],
            'return_value._category_slugs.return_value': ['foo', 'bar'],
        }
        with patch('redis_metrics.forms.R', **k) as mock_R:
            data = {'category_name': 'Foo', 'metrics': ['foo', 'bar']}
            form = MetricCategoryForm(data)
            self.assertTrue(form.is_valid())
            form.categorize_metrics()
            # This is what should happen in the form when POSTing
            mock_R.assert_has_calls([
                # happens in __init__
                call(),
                call().metric_slugs(),

                # happens in categorize_metrics
                call().reset_category('Foo', ['foo', 'bar'])
            ])



from __future__ import unicode_literals
from django import forms
from .models import R


class AggregateMetricForm(forms.Form):
    metrics = forms.MultipleChoiceField(
        choices=[],
        widget=forms.widgets.CheckboxSelectMultiple(),
    )

    def __init__(self, *args, **kwargs):
        super(AggregateMetricForm, self).__init__(*args, **kwargs)
        r = R()
        choices = [(slug, slug) for slug in r.metric_slugs()]
        self.fields['metrics'].choices = choices


class MetricCategoryForm(forms.Form):
    category_name = forms.CharField()
    metrics = forms.MultipleChoiceField(
        choices=[],
        widget=forms.widgets.CheckboxSelectMultiple(),
        required=False  # Don't require this, unchecking all metrics will
                        # remove a Category from Redis.
    )

    def __init__(self, *args, **kwargs):
        """See if this form is created with an initial Category. If so, we call
        out to ``R._category_slugs`` to find the pre-selected metrics. Any
        initial data should look like this:

            {'category_name': 'Whatever'}


        """
        super(MetricCategoryForm, self).__init__(*args, **kwargs)
        self.r = R()  # Keep a connection to our Redis wrapper
        # The list of available choices should include all metrics
        choices = [(slug, slug) for slug in self.r.metric_slugs()]
        self.fields['metrics'].choices = choices

        # If a "category" is provided, set inital values (pre-selected)
        initial = kwargs.get('initial', None)
        if initial and 'category_name' in initial:
            category = initial['category_name']
            self.fields['category_name'].initial = category
            self.fields['metrics'].initial = self.r._category_slugs(category)

    def categorize_metrics(self):
        """Called only on a valid form, this method will place the chosen
        metrics in the given catgory."""
        category = self.cleaned_data['category_name']
        metrics = self.cleaned_data['metrics']
        self.r.reset_category(category, metrics)



from __future__ import unicode_literals
import json
from django import template
from django.template.defaultfilters import stringfilter, mark_safe

register = template.Library()


@register.filter(name="json")
def to_json(values):
    return mark_safe(json.dumps(values))


@register.filter(name="int_list")
def to_int_list(values):
    """Converts the given list of vlues into a list of integers. If the
    integer conversion fails (e.g. non-numeric strings or None-values), this
    filter will include a 0 instead."""
    results = []
    for v in values:
        try:
            results.append(int(v))
        except (TypeError, ValueError):
            results.append(0)
    return results


@register.filter(name="int")
def to_int(value):
    """Converts the given string value into an integer. Returns 0 if the
    conversion fails."""
    try:
        return int(value)
    except (TypeError, ValueError):
        return 0


@register.filter(name='strip_metric_prefix')
@stringfilter
def strip_metric_prefix(value):
    """Strips the "m:<slug>" prefix from a metric.
    Applying this filter to the keys for each metric will have the following
    results:

    * Seconds -- from: ``m:<slug>:s:<yyyy-mm-dd-hh-MM-SS>`` to ``<yyyy-mm-dd-hh-MM-SS>``
    * Minutes -- from: ``m:<slug>:i:<yyyy-mm-dd-hh-MM>`` to ``<yyyy-mm-dd-hh-MM>``
    * Hourly -- from: ``m:<slug>:h:<yyyy-mm-dd-hh>`` to ``<yyyy-mm-dd-hh>``
    * Daily -- from: ``m:<slug>:<yyyy-mm-dd>`` to ``<yyyy-mm-dd>``
    * Weekly -- from ``m:<slug>:w:<num>`` to ``w:<num>``
    * Monthly -- from ``m:<slug>:m:<yyyy-mm>`` to ``m:<yyyy-mm>``
    * Yearly -- from ``m:<slug>:y:<yyyy>`` to ``y:<yyyy>``

    """
    return ':'.join(value.split(":")[2:])


@register.filter(name='metric_slug')
@stringfilter
def metric_slug(value):
    """Given a redis key value for a metric, returns only the slug.
    Applying this filter to the keys for each metric will have the following
    results:

    * Converts ``m:foo:s:<yyyy-mm-dd-hh-MM-SS>`` to ``foo``
    * Converts ``m:foo:i:<yyyy-mm-dd-hh-MM>`` to ``foo``
    * Converts ``m:foo:h:<yyyy-mm-dd-hh>`` to ``foo``
    * Converts ``m:foo:<yyyy-mm-dd>`` to ``foo``
    * Converts ``m:foo:w:<num>`` to ``foo``
    * Converts ``m:foo:m:<yyyy-mm>`` to ``foo``
    * Converts ``m:foo:y:<yyyy>`` to ``foo``

    """
    return value.split(":")[1]



"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""
from __future__ import unicode_literals
from datetime import datetime, timedelta
try:
    from unittest.mock import call, patch, Mock
except ImportError:
    from mock import call, patch, Mock

from django.test import TestCase
from django.test.utils import override_settings

from ..models import R, dedupe


TEST_SETTINGS = {
    'HOST': 'localhost',
    'PORT': 6379,
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
    'MIN_GRANULARITY': 'seconds',
    'MAX_GRANULARITY': 'yearly',
    'MONDAY_FIRST_DAY_OF_WEEK': False,
    'USE_ISO_WEEK_NUMBER': False,
}


@override_settings(REDIS_METRICS=TEST_SETTINGS)
class TestR(TestCase):
    """Tests for the ``R`` class."""

    def setUp(self):
        # Patch the connection to redis, but keep a reference to the
        # created StrictRedis instance, so we can make assertions about how
        # it's called.
        self.redis_patcher = patch('redis_metrics.models.redis.StrictRedis')
        mock_StrictRedis = self.redis_patcher.start()
        self.redis = mock_StrictRedis.return_value
        self.r = R()

    def tearDown(self):
        self.redis = self.redis_patcher.stop()
        super(TestR, self).tearDown()

    def test_dedupe(self):
        """Test the redis_metrics.models.dedupe function."""
        self.assertEqual(
            list(dedupe(['a', 'a', 'b', 'c', 'b', 'c', 'c', 'c'])),
            ['a', 'b', 'c']
        )

    def test__init__(self):
        """Test creation of an R object with parameters."""
        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            kwargs = {
                'decode_responses': True,
                'categories_key': 'CAT',
                'metric_slugs_key': 'MSK',
                'gauge_slugs_key': 'GSK',
                'host': 'HOST',
                'port': 'PORT',
                'db': 'DB',
                'password': 'PASSWORD',
                'socket_timeout': 1,
                'connection_pool': 1
            }
            inst = R(**kwargs)
            self.assertEqual(inst.host, "HOST")
            self.assertEqual(inst.port, "PORT")
            self.assertEqual(inst.db, "DB")
            self.assertEqual(inst.password, 'PASSWORD')
            self.assertEqual(inst.socket_timeout, 1)
            self.assertEqual(inst.connection_pool, 1)
            self.assertEqual(inst._categories_key, "CAT")
            self.assertEqual(inst._metric_slugs_key, "MSK")
            self.assertEqual(inst._gauge_slugs_key, "GSK")
            mock_redis.assert_called_once_with(
                decode_responses=True,
                host='HOST',
                port='PORT',
                db='DB',
                password="PASSWORD",
                socket_timeout=1,
                connection_pool=1
            )

    def test__init__with_default_kwargs(self):
        """Test creation of an R object without parameters."""
        r_kwargs = {
            'decode_responses': True,
            'host': 'localhost',
            'db': 0,
            'port': 6379,
            'password': None,
            'connection_pool': None,
            'socket_timeout': None
        }
        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            inst = R()
            self.assertEqual(inst.host, 'localhost')
            self.assertEqual(inst.port, 6379)
            self.assertEqual(inst.db, 0)
            self.assertEqual(inst.password, None)
            self.assertEqual(inst.socket_timeout, None)
            self.assertEqual(inst.connection_pool, None)
            self.assertEqual(inst._categories_key, 'categories')
            self.assertEqual(inst._metric_slugs_key, "metric-slugs")
            self.assertEqual(inst._gauge_slugs_key, "gauge-slugs")
            mock_redis.assert_called_once_with(**r_kwargs)

    def test__date_range(self):
        """Tests ``R._date_range`` at various granularities *without* a
        ``since`` date."""
        # minutes, seconds and hours should be capped at 300, 480, and 720
        d = datetime(2014, 1, 1)
        self.assertEqual(len(list(self.r._date_range('seconds', d))), 300)
        self.assertEqual(len(list(self.r._date_range('minutes', d))), 480)
        self.assertEqual(len(list(self.r._date_range('hourly', d))), 720)

        # Everything else should have a timedelta of 7 days
        self.assertEqual(len(list(self.r._date_range('daily', None))), 8)
        self.assertEqual(len(list(self.r._date_range('weekly', None))), 8)
        self.assertEqual(len(list(self.r._date_range('monthly', None))), 8)
        self.assertEqual(len(list(self.r._date_range('yearly', None))), 8)

        # Test with a specified `to` argument.
        since = datetime(2014, 1, 1)
        to = datetime(2015, 1, 1)
        self.assertEqual(len(list(self.r._date_range('daily', since, to))), 366)
        self.assertEqual(len(list(self.r._date_range('weekly', since, to))), 366)
        self.assertEqual(len(list(self.r._date_range('monthly', since, to))), 366)
        self.assertEqual(len(list(self.r._date_range('yearly', since, to))), 366)

    def test__date_range_seconds(self):
        """Tests ``R._date_range`` at the "seconds" granularity."""
        since = datetime.utcnow() - timedelta(seconds=5)
        values = self.r._date_range('seconds', since)
        self.assertEqual(len(list(values)), 5)

    def test__date_range_minutes(self):
        """Tests ``R._date_range`` at the "minutes" granularity."""
        since = datetime.utcnow() - timedelta(minutes=5)
        values = self.r._date_range('minutes', since)
        self.assertEqual(len(list(values)), 5)

    def test__date_range_hourly(self):
        """Tests ``R._date_range`` at the "hourly" granularity."""
        since = datetime.utcnow() - timedelta(hours=5)
        values = self.r._date_range('hourly', since)
        self.assertEqual(len(list(values)), 5)

    def test__date_range_daily(self):
        """Tests ``R._date_range`` at the "daily" granularity."""
        since = datetime.utcnow() - timedelta(days=5)
        values = self.r._date_range('daily', since)
        self.assertEqual(len(list(values)), 6)  # 6 because we do days + 1

    def test__date_range_weekly(self):
        """Tests ``R._date_range`` at the "weekly" granularity."""
        since = datetime.utcnow() - timedelta(days=5)
        values = self.r._date_range('weekly', since)
        self.assertEqual(len(list(values)), 6)  # 6 because we do days + 1

    def test__date_range_monthly(self):
        """Tests ``R._date_range`` at the "monthly" granularity."""
        since = datetime.utcnow() - timedelta(days=5)
        values = self.r._date_range('montly', since)
        self.assertEqual(len(list(values)), 6)  # 6 because we do days + 1

    def test__date_range_yearly(self):
        """Tests ``R._date_range`` at the "yearly" granularity."""
        since = datetime.utcnow() - timedelta(days=5)
        values = self.r._date_range('yearly', since)
        self.assertEqual(len(list(values)), 6)  # 6 because we do days + 1

    def test_categories(self):
        """Verify that ``R.categories()`` calls redis SMEMBERS command."""
        self.r.categories()
        self.redis.smembers.assert_called_once_with("categories")

    def test__category_key(self):
        """Creates a redis key for a given category string."""
        self.assertEqual(
            self.r._category_key("Sample Category"),
            u"c:Sample Category"
        )

    def test__category_slugs(self):
        """Verify that this returns an empty list or a list of slugs."""
        # When there are no results from redis
        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            mock_redis.return_value.smembers.return_value = set([])
            r = R()
            result = r._category_slugs("Sample Category")
            self.assertEqual(len(result), 0)

        # When there are no results from redis
        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            mock_redis.return_value.smembers.return_value = set(["slug-a", "slug-b"])
            r = R()
            result = r._category_slugs("Sample Category")
            self.assertEqual(set(result), set(['slug-a', 'slug-b']))

    @patch.object(R, '_category_slugs')
    def test__categorize(self, mock_category_slugs):
        """Categorizing a slug should add the correct key/values to Redis,
        and it should store the Category in a redis set."""

        # Sample category and metric slug
        cat = "Sample Category"
        cat_key = "c:Sample Category"
        slug = "sample-slug"

        with patch('redis_metrics.models.redis.StrictRedis') as mock_redis:
            redis_instance = mock_redis.return_value
            r = R()

            r._categorize(slug, cat)
            calls = [call.sadd(cat_key, slug), call.sadd("categories", cat)]
            redis_instance.assert_has_calls(calls, any_order=True)

    def test__granularities(self):
        """Tests ``R._granularities with default test settings."""
        # With default settings, min/max granularity is "seconds" and "yearly"
        self.assertEqual(
            list(self.r._granularities()),
            ['seconds', 'minutes', 'hourly', 'daily', 'weekly', 'monthly', 'yearly']
        )

    def test__granularities_with_altered_min(self):
        """Tests ``R._granularities with an altered minimum value."""
        test_settings = TEST_SETTINGS.copy()
        test_settings['MIN_GRANULARITY'] = 'daily'
        with override_settings(REDIS_METRICS=test_settings):
            self.assertEqual(
                list(self.r._granularities()),
                ['daily', 'weekly', 'monthly', 'yearly']
            )

    def test__granularities_with_altered_max(self):
        """Tests ``R._granularities with an altered maximum value."""
        test_settings = TEST_SETTINGS.copy()
        test_settings['MAX_GRANULARITY'] = 'daily'
        with override_settings(REDIS_METRICS=test_settings):
            self.assertEqual(
                list(self.r._granularities()),
                ['seconds', 'minutes', 'hourly', 'daily']
            )

    def test_get_metric_key_pattern_seconds(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('seconds', slug, date),
            "m:test-slug:s:2019-01-02-03-04-05"
        )

    def test_get_metric_key_pattern_minutes(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('minutes', slug, date),
            "m:test-slug:i:2019-01-02-03-04"
        )

    def test_get_metric_key_pattern_hourly(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('hourly', slug, date),
            "m:test-slug:h:2019-01-02-03"
        )

    def test_get_metric_key_pattern_daily(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('daily', slug, date),
            "m:test-slug:2019-01-02"
        )

    def test_get_metric_key_pattern_weekly_sunday_as_first(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['MONDAY_FIRST_DAY_OF_WEEK'] = False
        with override_settings(REDIS_METRICS=test_settings):
            slug = "test-slug"
            date = datetime(2012, 4, 1, 3, 4, 5)
            self.assertEqual(
                self.r._get_metric_key_pattern('weekly', slug, date),
                "m:test-slug:w:2012-14"
            )

    def test_get_metric_key_pattern_weekly_monday_at_first(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['MONDAY_FIRST_DAY_OF_WEEK'] = True
        with override_settings(REDIS_METRICS=test_settings):
            slug = "test-slug"
            date = datetime(2012, 4, 1, 3, 4, 5)
            self.assertEqual(
                self.r._get_metric_key_pattern('weekly', slug, date),
                "m:test-slug:w:2012-13"
            )

    def test_get_metric_key_pattern_weekly_iso_week(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['USE_ISO_WEEK_NUMBER'] = True
        with override_settings(REDIS_METRICS=test_settings):
            slug = "test-slug"
            date = datetime(2019, 1, 2, 3, 4, 5)
            self.assertEqual(
                self.r._get_metric_key_pattern('weekly', slug, date),
                "m:test-slug:w:2019-1"
            )

    def test_get_metric_key_pattern_monthly(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('monthly', slug, date),
            "m:test-slug:m:2019-01"
        )

    def test_get_metric_key_pattern_yearly(self):
        slug = "test-slug"
        date = datetime(2019, 1, 2, 3, 4, 5)
        self.assertEqual(
            self.r._get_metric_key_pattern('yearly', slug, date),
            "m:test-slug:y:2019"
        )

    def test__build_keys(self):
        """Tests ``R._build_keys``. with default arguments."""
        with patch('redis_metrics.models.datetime') as mock_datetime:
            mock_datetime.utcnow.return_value = datetime(2014, 7, 2, 12, 6, 34)
            expected_results = [
                "m:test-slug:s:2014-07-02-12-06-34",
                "m:test-slug:i:2014-07-02-12-06",
                "m:test-slug:h:2014-07-02-12",
                "m:test-slug:2014-07-02",
                "m:test-slug:w:2014-26",
                "m:test-slug:m:2014-07",
                "m:test-slug:y:2014",
            ]
            keys = self.r._build_keys('test-slug')
            self.assertEqual(keys, expected_results)

    def test__build_keys_seconds(self):
        """Tests ``R._build_keys``. with a *seconds* granularity."""
        d = datetime(2012, 4, 1, 11, 30, 59)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='seconds')
        self.assertEqual(keys, ['m:test-slug:s:2012-04-01-11-30-59'])

    def test__build_keys_minutes(self):
        """Tests ``R._build_keys``. with a *minutes* granularity."""
        d = datetime(2012, 4, 1, 11, 30)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='minutes')
        self.assertEqual(keys, ['m:test-slug:i:2012-04-01-11-30'])

    def test__build_keys_hourly(self):
        """Tests ``R._build_keys``. with a *hourly* granularity."""
        d = datetime(2012, 4, 1, 11, 30)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='hourly')
        self.assertEqual(keys, ['m:test-slug:h:2012-04-01-11'])

    def test__build_keys_daily(self):
        """Tests ``R._build_keys``. with a *daily* granularity."""
        d = datetime(2012, 4, 1)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='daily')
        self.assertEqual(keys, ['m:test-slug:2012-04-01'])

    def test__build_keys_weekly(self):
        """Tests ``R._build_keys``. with a *weekly* granularity."""
        d = datetime(2012, 4, 1)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='weekly')
        self.assertEqual(keys, ['m:test-slug:w:2012-14'])

    def test__build_keys_weekly_week_starts_monday(self):
        """Tests ``R._build_keys``. with a *weekly* granularity."""
        test_settings = TEST_SETTINGS.copy()
        test_settings['MONDAY_FIRST_DAY_OF_WEEK'] = True
        with override_settings(REDIS_METRICS=test_settings):
            d = datetime(2012, 4, 1)  # April Fools!
            keys = self.r._build_keys('test-slug', date=d, granularity='weekly')
            self.assertEqual(keys, ['m:test-slug:w:2012-13'])

    def test__build_keys_weekly_iso_week(self):
        """Tests ``R._build_keys``. with a *weekly* granularity."""
        test_settings = TEST_SETTINGS.copy()
        test_settings['USE_ISO_WEEK_NUMBER'] = True
        with override_settings(REDIS_METRICS=test_settings):
            d1 = datetime(2018, 12, 30)
            keys = self.r._build_keys('test-slug', date=d1, granularity='weekly')
            self.assertEqual(keys, ['m:test-slug:w:2018-52'])
            d2 = datetime(2018, 12, 31)
            keys = self.r._build_keys('test-slug', date=d2, granularity='weekly')
            self.assertEqual(keys, ['m:test-slug:w:2019-1'])
            d3 = datetime(2019, 1, 1)
            keys = self.r._build_keys('test-slug', date=d3, granularity='weekly')
            self.assertEqual(keys, ['m:test-slug:w:2019-1'])

    def test__build_keys_monthly(self):
        """Tests ``R._build_keys``. with a *monthly* granularity."""
        d = datetime(2012, 4, 1)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='monthly')
        self.assertEqual(keys, ['m:test-slug:m:2012-04'])

    def test__build_keys_yearly(self):
        """Tests ``R._build_keys``. with a *yearly* granularity."""
        d = datetime(2012, 4, 1)  # April Fools!
        keys = self.r._build_keys('test-slug', date=d, granularity='yearly')
        self.assertEqual(keys, ['m:test-slug:y:2012'])

    def test_metric_slugs(self):
        """Test that ``R.metric_slugs`` makes a call to Redis SMEMBERS."""
        self.r.metric_slugs()
        self.redis.assert_has_calls([call.smembers(self.r._metric_slugs_key)])

    @patch.object(R, '_category_slugs')
    def test_metric_slugs_by_category(self, mock_category_slugs):
        """Test that we get metric slugs organized by category."""
        # set up a mock return value for the category's redis set
        self.redis.smembers.return_value = ['Sample Category']
        # set up return value for mock call to _category_slugs
        mock_category_slugs.return_value = ['foo', 'bar']

        # Since this method all calls out to ``R.metric_slugs``, let's patch
        # that so it returns a list of all slugs, one of which will not be
        # categorized
        self.r.metric_slugs = Mock(return_value=['foo', 'bar', 'baz'])

        # Now, test the thing.
        result = self.r.metric_slugs_by_category()
        expected_result = {
            'Sample Category': ['foo', 'bar'],
            'Uncategorized': ['baz']
        }
        self.assertEqual(result, expected_result)
        self.redis.smembers.assert_called_once_with("categories")
        mock_category_slugs.assert_called_once_with("Sample Category")
        self.r.metric_slugs.assert_called_once_with()

    def test_delete_metric(self):
        """Verify that ``R.delete_metric`` deletes all keys and removes keys
        from the set of metric slugs."""

        # Make sure KEYS returns some data
        self.redis.keys.return_value = ["m:slug:0", "m:slug:1"]
        self.r.delete_metric('slug')  # call delete_metric

        # Verify that the metric data is removed as are the keys from the set
        self.redis.assert_has_calls([
            call.keys("m:slug:*"),
            call.delete("m:slug:0", "m:slug:1"),
            call.srem(self.r._metric_slugs_key, "slug")
        ])

    @patch.object(R, '_build_keys')
    def test_set_metric(self, mock_build_keys):
        """Test setting metrics using ``R.set_metric``."""
        # Define redis keys
        mock_build_keys.return_value = [
            'm:test-slug:s:2000-01-02-11-45-30',
            'm:test-slug:i:2000-01-02-11-45',
            'm:test-slug:h:2000-01-02-11',
            'm:test-slug:2000-01-02',
            'm:test-slug:m:2000-01',
            'm:test-slug:w:2000-01',
            'm:test-slug:y:2000'
        ]
        slug = 'test-slug'
        value = 42

        # get the metric keys so we can check for the appropriate calls
        self.r.set_metric(slug, value)

        # Verify that setting a metric adds the appropriate slugs to the keys
        # set and then incrememts each key
        self.redis.sadd.assert_called_once_with('metric-slugs', slug)
        self.redis.mset.assert_called_once_with({
            'm:test-slug:s:2000-01-02-11-45-30': 42,
            'm:test-slug:i:2000-01-02-11-45': 42,
            'm:test-slug:h:2000-01-02-11': 42,
            'm:test-slug:2000-01-02': 42,
            'm:test-slug:m:2000-01': 42,
            'm:test-slug:w:2000-01': 42,
            'm:test-slug:y:2000': 42,
        })

        # Expiration should not have gotten called
        self.assertFalse(self.redis.expire.called)

    @patch.object(R, '_build_keys')
    def test_set_metric_with_expiration(self, mock_build_keys):
        """Test setting metrics using ``R.set_metric`` with an expiration."""
        # Define redis keys
        mock_build_keys.return_value = [
            'm:test-slug:s:2000-01-02-11-45-30',
            'm:test-slug:i:2000-01-02-11-45',
            'm:test-slug:h:2000-01-02-11',
            'm:test-slug:2000-01-02',
            'm:test-slug:m:2000-01',
            'm:test-slug:w:2000-01',
            'm:test-slug:y:2000'
        ]
        slug = 'test-slug'
        value = 42

        # get the metric keys so we can check for the appropriate calls
        self.r.set_metric(slug, value, expire=500)

        # Verify that each key had an expiration set.
        self.redis.expire.assert_has_calls([
            call('m:test-slug:s:2000-01-02-11-45-30', 500),
            call('m:test-slug:i:2000-01-02-11-45', 500),
            call('m:test-slug:h:2000-01-02-11', 500),
            call('m:test-slug:2000-01-02', 500),
            call('m:test-slug:m:2000-01', 500),
            call('m:test-slug:w:2000-01', 500),
            call('m:test-slug:y:2000', 500),
        ])

    @patch.object(R, '_categorize')
    def test_set_metric_with_category(self, mock_categorize):
        """Test setting metrics using ``R.set_metric`` when a category
        is provided."""

        # get the metric keys so we can check for the appropriate calls
        self.r.set_metric('test-slug', 42, category='Test Category')
        mock_categorize.assert_called_once_with('test-slug', "Test Category")

    def test_metric(self):
        """Test setting metrics using ``R.metric``."""

        slug = 'test-metric'
        n = 1

        # get the keys used for the metric, so we can check for the appropriate
        # calls
        keys = self.r._build_keys(slug)
        second, minute, hour, day, week, month, year = keys
        self.r.metric(slug, num=n)

        # Verify that setting a metric adds the appropriate slugs to the keys
        # set and then incrememts each key
        self.redis.assert_has_calls([
            call.sadd(self.r._metric_slugs_key, slug),
            call.pipeline(),
            call.pipeline().incr(second, n),
            call.pipeline().incr(minute, n),
            call.pipeline().incr(hour, n),
            call.pipeline().incr(day, n),
            call.pipeline().incr(week, n),
            call.pipeline().incr(month, n),
            call.pipeline().incr(year, n),
        ])

        # Expiration should not have gotten called
        self.assertFalse(self.redis.expire.called)

    def test_metric_with_overridden_granularities(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['MIN_GRANULARITY'] = 'daily'
        test_settings['MAX_GRANULARITY'] = 'weekly'
        with override_settings(REDIS_METRICS=test_settings):
            slug = 'test-metric'
            n = 1

            # get the keys used for the metric, so we can check for the appropriate
            # calls
            daily, weekly = self.r._build_keys(slug)
            self.r.metric(slug, num=n)

            # Verify that setting a metric adds the appropriate slugs to the keys
            # set and then incrememts each key
            self.redis.assert_has_calls([
                call.sadd(self.r._metric_slugs_key, slug),
                call.pipeline(),
                call.pipeline().incr(daily, n),
                call.pipeline().incr(weekly, n),
            ])

            # Expiration should not have gotten called
            self.assertFalse(self.redis.expire.called)

    @patch.object(R, '_categorize')
    def test_metric_with_category(self, mock_categorize):
        """The ``metric`` method should call ``_categorize`` if passed a
        ``category`` argument."""
        category = "Some Category"
        slug = 'categorized-metric'
        n = 1

        # get the keys used for the metric, so we can check for calls
        keys = self.r._build_keys(slug)
        second, minute, hour, day, week, month, year = keys
        self.r.metric(slug, num=n, category=category)

        # Verify that setting a metric adds the appropriate slugs to the keys
        # set and then incrememts each key
        self.redis.assert_has_calls([
            call.sadd(self.r._metric_slugs_key, slug),
            call.pipeline(),
            call.pipeline().incr(second, n),
            call.pipeline().incr(minute, n),
            call.pipeline().incr(hour, n),
            call.pipeline().incr(day, n),
            call.pipeline().incr(week, n),
            call.pipeline().incr(month, n),
            call.pipeline().incr(year, n),
        ])

        # Make sure this gets categorized.
        mock_categorize.assert_called_once_with(slug, category)

        # Expiration should not have gotten called
        self.assertFalse(self.redis.expire.called)

    @patch.object(R, '_categorize')
    def test_metric_with_expiration(self, mock_categorize):
        """The ``metric`` method should call the redis ``expire`` method if
        passed an ``expire`` argument."""

        slug = 'categorized-metric'
        n = 1

        # get the keys used for the metric, so we can check for calls
        keys = self.r._build_keys(slug)
        self.r.metric(slug, num=n, expire=3600)

        # Verify that setting a metric adds the appropriate slugs to the keys
        # set and then incrememts each key
        call_list = [call.sadd(self.r._metric_slugs_key, slug), call.pipeline()]
        for k in keys:
            call_list.append(call.pipeline().incr(k, n))
            call_list.append(call.pipeline().expire(k, 3600))

        self.redis.assert_has_calls(call_list)

        # Make sure nothing was categorized.
        self.assertFalse(mock_categorize.called)

    def test_get_metric(self):
        """Tests getting a single metric; ``R.get_metric``."""
        slug = 'test-metric'
        self.r.get_metric(slug)

        # Verify that we GET the keys from redis
        sec, min, hour, day, week, month, year = self.r._build_keys(slug)
        self.redis.assert_has_calls([
            call.get(sec),
            call.get(min),
            call.get(hour),
            call.get(day),
            call.get(week),
            call.get(month),
            call.get(year),
        ])

    def test_get_metric_with_overridden_granularities(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['MIN_GRANULARITY'] = 'daily'
        test_settings['MAX_GRANULARITY'] = 'weekly'
        with override_settings(REDIS_METRICS=test_settings):
            slug = 'test-metric'
            self.r.get_metric(slug)

            # Verify that we GET the keys from redis
            day, week = self.r._build_keys(slug)
            self.redis.assert_has_calls([
                call.get(day),
                call.get(week),
            ])

    def test_get_metrics(self):
        # Set a return value for mget, so all of the method gets exercised.
        prev_return = self.redis.mget.return_value
        self.redis.mget.return_value = ['1', '2']

        # Slugs for metrics we want
        slugs = ['metric-1', 'metric-2']

        # Build the various keys for each metric
        keys_list = []
        for s in slugs:
            keys_list.append(self.r._build_keys(s))

        # construct the calls to redis.mget
        calls = [call(*keys) for keys in keys_list]

        # Test our method
        self.r.get_metrics(slugs)
        self.assertEqual(self.redis.mget.call_args_list, calls)
        self.assertTrue(self.redis.mget.called)  # mget was called...
        self.assertEqual(self.redis.mget.call_count, 2)  # ...twice

        # Reset mget's previous return value
        self.redis.mget.return_value = prev_return

    def test_get_metrics_with_overridden_granularities(self):
        test_settings = TEST_SETTINGS.copy()
        test_settings['MIN_GRANULARITY'] = 'daily'
        test_settings['MAX_GRANULARITY'] = 'weekly'
        with override_settings(REDIS_METRICS=test_settings):
            # Set a return value for mget, so all of the method gets exercised.
            prev_return = self.redis.mget.return_value
            self.redis.mget.return_value = ['1', '2']

            # Slugs for metrics we want
            slugs = ['metric-1', 'metric-2']

            # Build the various keys for each metric
            keys_list = []
            for s in slugs:
                keys_list.append(self.r._build_keys(s))

            # construct the calls to redis.mget
            calls = [call(*keys) for keys in keys_list]

            # Test our method
            self.r.get_metrics(slugs)
            self.assertEqual(self.redis.mget.call_args_list, calls)
            self.assertTrue(self.redis.mget.called)  # mget was called...
            self.assertEqual(self.redis.mget.call_count, 2)  # ...twice

            # Reset mget's previous return value
            self.redis.mget.return_value = prev_return

    def test_get_category_metrics(self):
        """returns metrics for a given category"""
        r = R()
        # Mock methods called by `get_category_metrics`
        r._category_slugs = Mock(return_value=['some-slug'])
        r.get_metrics = Mock(return_value="RESULT")
        results = r.get_category_metrics("Sample Category")
        self.assertEqual(results, 'RESULT')
        r._category_slugs.assert_called_once_with("Sample Category")
        r.get_metrics.assert_called_once_with(['some-slug'])

    def test_delete_category(self):
        r_kwargs = {
            'decode_responses': True,
            'host': 'localhost',
            'db': 0,
            'port': 6379,
            'password': None,
            'connection_pool': None,
            'socket_timeout': None
        }
        with patch("redis_metrics.models.redis.StrictRedis") as mock_redis:
            r = R()
            r.delete_category("Foo")
            mock_redis.assert_has_calls([
                call(**r_kwargs),
                call().delete('c:Foo'),
                call().srem("categories", "Foo")
            ])

    @patch.object(R, "delete_category")
    def test_reset_category_with_no_metrics(self, mock_delete_category):
        """Calling ``reset_category`` with an empty list of metrics should
        just delete the category."""
        with patch("redis_metrics.models.redis.StrictRedis"):
            r = R()
            r.reset_category("Stuff", [])
            mock_delete_category.assert_called_once_with("Stuff")

    def test_reset_category(self):
        r_kwargs = {
            'decode_responses': True,
            'host': 'localhost',
            'db': 0,
            'port': 6379,
            'password': None,
            'connection_pool': None,
            'socket_timeout': None
        }
        with patch("redis_metrics.models.redis.StrictRedis") as mock_redis:
            r = R()
            r.reset_category("Stuff", ['foo', 'bar'])
            mock_redis.assert_has_calls([
                call(**r_kwargs),
                call().sadd('c:Stuff', 'foo', 'bar'),
                call().sadd('categories', 'Stuff'),
            ])

    def _metric_history_keys(self, slugs, since=None, to=None,
                             granularity='daily'):
        """generates the same list of keys used in ``get_metric_history``.
        These can then be used to test for calls to redis. Note: This is
        duplicate code from ``get_metric_history`` :-/ """
        if type(slugs) != list:
            slugs = [slugs]
        keys = []
        for slug in slugs:
            for date in self.r._date_range(granularity, since, to):
                keys += self.r._build_keys(slug, date, granularity)
        keys = list(dedupe(keys))
        return keys

    def _test_get_metric_history(self, slugs, since=None, to=None,
                                 granularity=None):
        """actual test code for ``R.get_metric_history``."""
        keys = self._metric_history_keys(slugs, since, to, granularity)
        self.r.get_metric_history(
            slugs, since=since, to=to, granularity=granularity)
        self.redis.assert_has_calls([call.mget(keys)])

    def test_get_metric_history_hourly(self):
        """Tests ``R.get_metric_history`` with hourly granularity."""
        self._test_get_metric_history('test-slug', granularity='hourly')

        # with specified to/since dates
        self._test_get_metric_history(
            'test-slug',
            since=datetime(2014, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='hourly'
        )

    def test_get_metric_history_daily(self):
        """Tests ``R.get_metric_history`` with daily granularity."""
        self._test_get_metric_history('test-slug', granularity='daily')

        # with specified to/since dates
        self._test_get_metric_history(
            'test-slug',
            since=datetime(2014, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='daily'
        )

    def test_get_metric_history_weekly(self):
        """Tests ``R.get_metric_history`` with weekly granularity."""
        self._test_get_metric_history('test-slug', granularity='weekly')

        # with specified to/since dates
        self._test_get_metric_history(
            'test-slug',
            since=datetime(2014, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='weekly'
        )

    def test_get_metric_history_monthly(self):
        """Tests ``R.get_metric_history`` with monthly granularity."""
        self._test_get_metric_history('test-slug', granularity='monthly')

        # with specified to/since dates
        self._test_get_metric_history(
            'test-slug',
            since=datetime(2014, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='monthly'
        )

    def test_get_metric_history_yearly(self):
        """Tests ``R.get_metric_history`` with yearly granularity."""
        self._test_get_metric_history('test-slug', granularity='yearly')

        # with specified to/since dates
        self._test_get_metric_history(
            'test-slug',
            since=datetime(2014, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='yearly'
        )

    def test_get_metric_multiple_history_hourly(self):
        self._test_get_metric_history(['foo', 'bar'], granularity='hourly')

        # with specified to/since dates
        self._test_get_metric_history(
            ['foo', 'bar'],
            since=datetime(2015, 1, 1),
            to=datetime(2015, 2, 1),
            granularity='hourly'
        )

    def test_get_metric_multiple_history_daily(self):
        self._test_get_metric_history(['foo', 'bar'], granularity='daily')

        # with specified to/since dates
        self._test_get_metric_history(
            ['foo', 'bar'],
            since=datetime(2015, 1, 1),
            to=datetime(2015, 2, 1),
            granularity='daily'
        )

    def test_get_metric_multiple_history_weekly(self):
        self._test_get_metric_history(['foo', 'bar'], granularity='weekly')

        # with specified to/since dates
        self._test_get_metric_history(
            ['foo', 'bar'],
            since=datetime(2015, 1, 1),
            to=datetime(2015, 2, 1),
            granularity='weekly'
        )

    def test_get_metric_multiple_history_monthly(self):
        self._test_get_metric_history(['foo', 'bar'], granularity='monthly')

        # with specified to/since dates
        self._test_get_metric_history(
            ['foo', 'bar'],
            since=datetime(2015, 1, 1),
            to=datetime(2015, 12, 1),
            granularity='monthly'
        )

    def test_get_metric_multiple_history_yearly(self):
        self._test_get_metric_history(['foo', 'bar'], granularity='yearly')

        # with specified to/since dates
        self._test_get_metric_history(
            ['foo', 'bar'],
            since=datetime(2015, 1, 1),
            to=datetime(2016, 1, 1),
            granularity='yearly'
        )

    @patch.object(R, '_date_range')
    def test_get_metric_history_replaces_none_with_zero(self, mock_date_range):
        """Ensure that None-values get replaced with Zeros in
        ``R.get_metric_history``."""

        # Mock the _date_range method so we can specify it's return values.
        mock_date_range.return_value = [
            datetime(2000, 1, 1),
            datetime(2000, 1, 2),
            datetime(2000, 1, 3),
            datetime(2000, 1, 4),
        ]

        # Temporarily change the return value for mget
        mget_return = self.redis.mget.return_value
        self.redis.mget.return_value = ['1', '2', None, '3']

        # Note: we're not providing a since parameter here, since we've
        # mocked the R._date_range method.
        results = self.r.get_metric_history("foo", granularity="daily")

        # Format the range of dates that for which we should get results
        expected = [
            ('m:foo:2000-01-01', '1'),
            ('m:foo:2000-01-02', '2'),
            ('m:foo:2000-01-03', 0),
            ('m:foo:2000-01-04', '3'),
        ]
        self.assertEqual(results, expected)

        # Reset mget's previous return value
        self.redis.mget.return_value = mget_return

    @patch.object(R, 'get_metric_history')
    def test_get_metric_history_as_columns(self, mock_metric_hist):
        # set up some sample (yearly) metrics
        mock_metric_hist.return_value = [
            ("m:bar:y:2012", '1'),
            ('m:bar:y:2013', '2'),
            ('m:foo:y:2012', '3'),
            ('m:foo:y:2013', '4'),
        ]
        expected_results = [
            ('Period', 'foo', 'bar'),
            ('y:2012', '3', '1'),
            ('y:2013', '4', '2'),
        ]
        with patch('redis_metrics.models.redis.StrictRedis'):
            r = R()
            kwargs = {
                'slugs': ['foo', 'bar'],
                'since': None,
                'granularity': 'yearly',
            }
            results = r.get_metric_history_as_columns(**kwargs)
            self.assertEqual(results, expected_results)

    def _test_get_metric_history_as_columns(self, slugs, granularity):
        """Test that R.get_metric_history_as_columns makes calls to the
        following functions:

        * ``R.r.mget``
        * ``R.get_metric_history``
        * ``templatetags.metric_slug``
        * ``templatetags.strip_metric_prefix``

        """
        keys = self._metric_history_keys(slugs, granularity=granularity)
        self.r.get_metric_history_as_columns(slugs, granularity=granularity)

        # Verifies the correct call to redis
        self.redis.assert_has_calls([call.mget(keys)])

        # Verify that the method gets called correctly
        with patch('redis_metrics.models.R') as mock_r:
            r = mock_r.return_value  # Get an instance of our Mocked R class
            r.get_metric_history_as_columns(slugs, granularity=granularity)
            mock_r.assert_has_calls([
                call().get_metric_history_as_columns(
                    slugs, granularity=granularity
                )
            ])

    def test_get_metric_history_as_columns_hourly(self):
        self._test_get_metric_history_as_columns(['foo', 'bar'], 'hourly')

    def test_get_metric_history_as_columns_daily(self):
        self._test_get_metric_history_as_columns(['foo', 'bar'], 'daily')

    def test_get_metric_history_as_columns_weekly(self):
        self._test_get_metric_history_as_columns(['foo', 'bar'], 'weekly')

    def test_get_metric_history_as_columns_monthly(self):
        self._test_get_metric_history_as_columns(['foo', 'bar'], 'monthly')

    def test_get_metric_history_as_columns_yearly(self):
        self._test_get_metric_history_as_columns(['foo', 'bar'], 'yearly')

    @patch.object(R, 'get_metric_history')
    def test_get_metric_history_chart_data(self, mock_metric_hist):
        # set up some sample (yearly) metrics
        mock_metric_hist.return_value = [
            ("m:bar:y:2012", '1'),
            ('m:bar:y:2013', '2'),
            ('m:foo:y:2012', '3'),
            ('m:foo:y:2013', '4'),
        ]
        expected_results = {
            'periods': ['y:2012', 'y:2013'],
            'data': [
                {'slug': 'bar', 'values': ['1', '2']},
                {'slug': 'foo', 'values': ['3', '4']},
            ]
        }
        with patch('redis_metrics.models.redis.StrictRedis'):
            r = R()
            kwargs = {
                'slugs': ['foo', 'bar'],
                'since': None,
                'granularity': 'yearly',
            }
            results = r.get_metric_history_chart_data(**kwargs)
            self.assertEqual(results, expected_results)

    def test_gauge_slugs(self):
        """Tests that ``R.gauge_slugs`` calls the SMEMBERS command."""
        self.r.gauge_slugs()
        self.redis.assert_has_calls([call.smembers(self.r._gauge_slugs_key)])

    def test__gauge_key(self):
        """Tests that ``R._gauge_key`` correctly generates gauge keys."""
        key = self.r._gauge_key('test-gauge')
        self.assertEqual(key, 'g:test-gauge')

    def test_gauge(self):
        """Tests setting a gauge with ``R.gauge``. Verifies that the gauge slug
        is added to the set of gauge slugs and that the value gets set."""
        self.r.gauge('test-gauge', 9000)
        self.redis.assert_has_calls([
            call.sadd(self.r._gauge_slugs_key, 'test-gauge'),
            call.set('g:test-gauge', 9000),
        ])

    def test_get_gauge(self):
        """Tests retrieving a gague with ``R.get_gauge``. Verifies that the
        Redis GET command is called with the correct key."""
        self.r.get_gauge('test-gauge')
        self.redis.assert_has_calls([call.get('g:test-gauge')])

    def test_delete_gauge(self):
        """Tests deltion of a gauge."""
        self.r.delete_gauge("test-gauge")
        self.redis.assert_has_calls([
            call.delete('g:test-gauge'),
            call.srem(self.r._gauge_slugs_key, "test-gauge"),
        ])



from __future__ import unicode_literals
from django.urls import re_path

from .views import (
    AggregateHistoryView, AggregateDetailView, AggregateFormView,
    CategoryFormView, DefaultView, GaugesView, MetricDetailView,
    MetricHistoryView, MetricsListView,
)


urlpatterns = [
    re_path(
        r'^categorize/(?P<category_name>.*)/$',
        CategoryFormView.as_view(),
        name='redis_metrics_categorize'
    ),
    re_path(
        r'^categorize/$',
        CategoryFormView.as_view(),
        name='redis_metrics_categorize'
    ),
    re_path(
        r'^aggregate/category/(?P<category>.*)/$',
        AggregateDetailView.as_view(),
        name='redis_metric_aggregate_detail_by_category'
    ),
    re_path(
        r'^aggregate/(?P<slugs>.*)/(?P<granularity>.*)/$',
        AggregateHistoryView.as_view(),
        name='redis_metric_aggregate_history'
    ),
    re_path(
        r'^aggregate/(?P<slugs>.*)/$',
        AggregateDetailView.as_view(),
        name='redis_metric_aggregate_detail'
    ),
    re_path(
        r'^aggregate/$',
        AggregateFormView.as_view(),
        name='redis_metric_aggregate'
    ),
    re_path(
        r'^list/$',
        MetricsListView.as_view(),
        name='redis_metrics_list'
    ),
    re_path(
        r'^gauges/$',
        GaugesView.as_view(),
        name='redis_metrics_gauges'
    ),
    re_path(
        r'^(?P<slug>.*)/(?P<granularity>.*)/$',
        MetricHistoryView.as_view(),
        name='redis_metric_history'
    ),
    re_path(
        r'^(?P<slug>.*)/$',
        MetricDetailView.as_view(),
        name='redis_metric_detail'
    ),
    re_path(
        r'^$',
        DefaultView.as_view(),
        name='redis_metrics_default'
    ),
]



from __future__ import unicode_literals
import random

from datetime import datetime, timedelta
from .models import R


_redis_model = None


def get_r():
    global _redis_model
    if not _redis_model:
        _redis_model = R()
    return _redis_model


def set_metric(slug, value, category=None, expire=None, date=None):
    """Create/Increment a metric."""
    get_r().set_metric(slug, value, category=category, expire=expire, date=date)


def metric(slug, num=1, category=None, expire=None, date=None):
    """Create/Increment a metric."""
    get_r().metric(slug, num=num, category=category, expire=expire, date=date)


def gauge(slug, current_value):
    """Set a value for a Gauge"""
    get_r().gauge(slug, current_value)


def generate_test_metrics(slug='test-metric', num=100, randomize=False,
                          cap=None, increment_value=100):
    """Generate some dummy metrics for the given ``slug``.

    * ``slug`` -- the Metric slug
    * ``num`` -- Number of days worth of metrics (default is 100)
    * ``randomize`` -- Generate random metric values (default is False)
    * ``cap`` -- If given, cap the maximum metric value.
    * ``increment_value`` -- The amount by which we increment metrics on
      subsequent days. If ``randomize`` is True, this value is used to
      generate a ceiling for random values.

    NOTE: This only generates metrics for daily and larger granularities.

    """
    r = get_r()
    i = 0
    if randomize:
        random.seed()

    r.r.sadd(r._metric_slugs_key, slug)  # Store the slug created.
    for date in r._date_range('daily', datetime.utcnow() - timedelta(days=num)):
        # Only keep the keys for daily and above granularities.
        keys = r._build_keys(slug, date=date)
        keys = [k for k in keys if k.split(":")[2] not in ['i', 's', 'h']]
        for key in keys:
            # The following is normally done in r.metric, but we're adding
            # metrics for past days here, so this is duplicate code.
            value = i
            if randomize:
                value = random.randint(0, i + increment_value)
            if cap and r.r.get(key) >= cap:
                value = 0  # Dont' increment this one any more.
            r.r.incr(key, value)
        i += increment_value


def delete_test_metrics(slug='test-metric', num=100):
    """Deletes the metrics created by ``generate_test_metrics``."""
    r = get_r()
    for date in r._date_range('daily', datetime.utcnow() - timedelta(days=num)):
        keys = r._build_keys(slug, date=date)
        r.r.srem(r._metric_slugs_key, slug)  # remove metric slugs
        r.r.delete(*keys)  # delete the metrics



from __future__ import unicode_literals
from datetime import datetime, timedelta

from django import template
from redis_metrics.utils import get_r
from redis_metrics.settings import GRANULARITIES

register = template.Library()


@register.inclusion_tag("redis_metrics/_since.html")
def metrics_since(slugs, years, link_type="detail", granularity=None):
    """Renders a template with a menu to view a metric (or metrics) for a
    given number of years.

    * ``slugs`` -- A Slug or a set/list of slugs
    * ``years`` -- Number of years to show past metrics
    * ``link_type`` -- What type of chart do we want ("history" or "aggregate")
        * history  -- use when displaying a single metric's history
        * aggregate -- use when displaying aggregate metric history
    * ``granularity`` -- For "history" only; show the metric's granularity;
      default is "daily"

    """
    now = datetime.utcnow()

    # Determine if we're looking at one slug or multiple slugs
    if type(slugs) in [list, set]:
        slugs = "+".join(s.lower().strip() for s in slugs)

    # Set the default granularity if it's omitted
    granularity = granularity.lower().strip() if granularity else "daily"

    # Each item is: (slug, since, text, granularity)
    # Always include values for Today, 1 week, 30 days, 60 days, 90 days...
    slug_values = [
        (slugs, now - timedelta(days=1), "Today", granularity),
        (slugs, now - timedelta(days=7), "1 Week", granularity),
        (slugs, now - timedelta(days=30), "30 Days", granularity),
        (slugs, now - timedelta(days=60), "60 Days", granularity),
        (slugs, now - timedelta(days=90), "90 Days", granularity),
    ]

    # Then an additional number of years
    for y in range(1, years + 1):
        t = now - timedelta(days=365 * y)
        text = "{0} Years".format(y)
        slug_values.append((slugs, t, text, granularity))
    return {'slug_values': slug_values, 'link_type': link_type.lower().strip()}


@register.inclusion_tag("redis_metrics/_gauge.html")
def gauge(slug, maximum=9000, size=200, coerce='float'):
    """Include a Donut Chart for the specified Gauge.

    * ``slug`` -- the unique slug for the Gauge.
    * ``maximum`` -- The maximum value for the gauge (default is 9000)
    * ``size`` -- The size (in pixels) of the gauge (default is 200)
    * ``coerce`` -- type to which gauge values should be coerced. The default
      is float. Use ``{% gauge some_slug coerce='int' %}`` to coerce to integer

    """
    coerce_options = {'float': float, 'int': int, 'str': str}
    coerce = coerce_options.get(coerce, float)

    redis = get_r()
    value = coerce(redis.get_gauge(slug))
    if value < maximum and coerce == float:
        diff = round(maximum - value, 2)
    elif value < maximum:
        diff = maximum - value
    else:
        diff = 0

    return {
        'slug': slug,
        'current_value': value,
        'max_value': maximum,
        'size': size,
        'diff': diff,
    }


@register.inclusion_tag("redis_metrics/_metric_list.html")
def metric_list():
    r = get_r()
    return {
        'metrics': r.metric_slugs_by_category(),
    }


@register.inclusion_tag("redis_metrics/_metric_detail.html")
def metric_detail(slug, with_data_table=False):
    """Template Tag to display a metric's *current* detail.

    * ``slug`` -- the metric's unique slug
    * ``with_data_table`` -- if True, prints the raw data in a table.

    """
    r = get_r()
    granularities = list(r._granularities())
    metrics = r.get_metric(slug)
    metrics_data = []
    for g in granularities:
        metrics_data.append((g, metrics[g]))

    return {
        'granularities': [g.title() for g in granularities],
        'slug': slug,
        'metrics': metrics_data,
        'with_data_table': with_data_table,
    }


@register.inclusion_tag("redis_metrics/_metric_history.html")
def metric_history(slug, granularity="daily", since=None, to=None,
                   with_data_table=False):
    """Template Tag to display a metric's history.

    * ``slug`` -- the metric's unique slug
    * ``granularity`` -- the granularity: daily, hourly, weekly, monthly, yearly
    * ``since`` -- a datetime object or a string string matching one of the
      following patterns: "YYYY-mm-dd" for a date or "YYYY-mm-dd HH:MM:SS" for
      a date & time.
    * ``to`` -- the date until which we start pulling metrics
    * ``with_data_table`` -- if True, prints the raw data in a table.

    """
    r = get_r()
    try:
        if since and len(since) == 10:  # yyyy-mm-dd
            since = datetime.strptime(since, "%Y-%m-%d")
        elif since and len(since) == 19:  # yyyy-mm-dd HH:MM:ss
            since = datetime.strptime(since, "%Y-%m-%d %H:%M:%S")

        if to and len(to) == 10:  # yyyy-mm-dd
            to = datetime.strptime(since, "%Y-%m-%d")
        elif to and len(to) == 19:  # yyyy-mm-dd HH:MM:ss
            to = datetime.strptime(to, "%Y-%m-%d %H:%M:%S")

    except (TypeError, ValueError):
        # assume we got a datetime object or leave since = None
        pass

    metric_history = r.get_metric_history(
        slugs=slug,
        since=since,
        to=to,
        granularity=granularity
    )

    return {
        'since': since,
        'to': to,
        'slug': slug,
        'granularity': granularity,
        'metric_history': metric_history,
        'with_data_table': with_data_table,
    }


@register.inclusion_tag("redis_metrics/_aggregate_detail.html")
def aggregate_detail(slug_list, with_data_table=False):
    """Template Tag to display multiple metrics.

    * ``slug_list`` -- A list of slugs to display
    * ``with_data_table`` -- if True, prints the raw data in a table.

    """
    r = get_r()
    metrics_data = []
    granularities = r._granularities()

    # XXX converting granularties into their key-name for metrics.
    keys = ['seconds', 'minutes', 'hours', 'day', 'week', 'month', 'year']
    key_mapping = {gran: key for gran, key in zip(GRANULARITIES, keys)}
    keys = [key_mapping[gran] for gran in granularities]

    # Our metrics data is of the form:
    #
    #   (slug, {time_period: value, ... }).
    #
    # Let's convert this to (slug, list_of_values) so that the list of
    # values is in the same order as the granularties
    for slug, data in r.get_metrics(slug_list):
        values = [data[t] for t in keys]
        metrics_data.append((slug, values))

    return {
        'chart_id': "metric-aggregate-{0}".format("-".join(slug_list)),
        'slugs': slug_list,
        'metrics': metrics_data,
        'with_data_table': with_data_table,
        'granularities': [g.title() for g in keys],
    }


@register.inclusion_tag("redis_metrics/_aggregate_history.html")
def aggregate_history(slugs, granularity="daily", since=None, with_data_table=False):
    """Template Tag to display history for multiple metrics.

    * ``slug_list`` -- A list of slugs to display
    * ``granularity`` -- the granularity: seconds, minutes, hourly,
                         daily, weekly, monthly, yearly
    * ``since`` -- a datetime object or a string string matching one of the
      following patterns: "YYYY-mm-dd" for a date or "YYYY-mm-dd HH:MM:SS" for
      a date & time.
    * ``with_data_table`` -- if True, prints the raw data in a table.

    """
    r = get_r()
    slugs = list(slugs)

    try:
        if since and len(since) == 10:  # yyyy-mm-dd
            since = datetime.strptime(since, "%Y-%m-%d")
        elif since and len(since) == 19:  # yyyy-mm-dd HH:MM:ss
            since = datetime.strptime(since, "%Y-%m-%d %H:%M:%S")
    except (TypeError, ValueError):
        # assume we got a datetime object or leave since = None
        pass

    history = r.get_metric_history_chart_data(
        slugs=slugs,
        since=since,
        granularity=granularity
    )

    return {
        'chart_id': "metric-aggregate-history-{0}".format("-".join(slugs)),
        'slugs': slugs,
        'since': since,
        'granularity': granularity,
        'metric_history': history,
        'with_data_table': with_data_table,
    }



"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""
from __future__ import unicode_literals
from datetime import datetime
try:
    from unittest.mock import call, patch
except ImportError:
    from mock import call, patch

from django.urls import reverse
from django.test import TestCase, Client
from django.test.utils import override_settings


TEST_SETTINGS = {
    'HOST': 'localhost',
    'PORT': 6379,
    'DB': 0,
    'PASSWORD': None,
    'SOCKET_TIMEOUT': None,
    'SOCKET_CONNECTION_POOL': None,
    'MIN_GRANULARITY': 'seconds',
    'MAX_GRANULARITY': 'yearly',
    'MONDAY_FIRST_DAY_OF_WEEK': False,
    'USE_ISO_WEEK_NUMBER': False,
}


@override_settings(REDIS_METRICS=TEST_SETTINGS)
class TestViews(TestCase):
    url = 'redis_metrics.urls'

    def _get_user_model(self):
        try:  # pragma: no cover
            # Django >= 1.5
            from django.contrib.auth import get_user_model  # pragma: no cover
            User = get_user_model()  # pragma: no cover
        except ImportError:  # pragma: no cover
            # Fallback for Django 1.4 (or lower)
            from django.contrib.auth.models import User  # pragma: no cover
        return User

    def setUp(self):
        # Patch the connection to redis, but keep a reference to the
        # created StrictRedis instance, so we can make assertions about how
        # it's called.
        self.redis_patcher = patch('redis_metrics.models.redis.StrictRedis')
        mock_StrictRedis = self.redis_patcher.start()
        self.redis = mock_StrictRedis.return_value

        User = self._get_user_model()
        self.user = User.objects.create_superuser(
            username="redis_metrics_test_user",
            email="redis_metrics_test_user@example.com",
            password="secret"
        )
        assert self.client.login(
            username="redis_metrics_test_user",
            password="secret")
        self.unauthed_client = Client()  # Keep an unauthenticated client

    def tearDown(self):
        self.redis_patcher.stop()
        self.user.delete()

    def assertUnauthedRequestRedirects(self, url):
        resp = self.unauthed_client.get(url)
        self.assertEqual(resp.status_code, 302)
        return resp

    def test_default_view(self):
        url = reverse('redis_metrics_default')
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, "redis_metrics/default.html")

    def test_gauges_view(self):
        """Test the ``GaugesView``."""
        url = reverse('redis_metrics_gauges')
        with patch('redis_metrics.views.get_r') as mock_r:
            # Set appropriate return values for methods that'll get called
            # in the MetricsListView.
            r = mock_r.return_value  # Get an instance of our Mocked R class
            r.gauge_slugs.return_value = set(['gauge-a', 'gauge-b'])

            # Do the Request and test for content
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('Gauges', str(resp.content))
            self.assertEqual(
                resp.context['gauges'],
                set(['gauge-a', 'gauge-b'])
            )

            # Make sure our Mock R object called the right methods.
            mock_r.assert_has_calls([call().gauge_slugs()])

    def test_metrics_list(self):
        """Test the ``MetricsListView``."""
        url = reverse('redis_metrics_list')
        with patch('redis_metrics.views.get_r') as mock_r:
            # Set appropriate return values for methods that'll get called
            # in the MetricsListView.
            r = mock_r.return_value  # Get an instance of our Mocked R class
            r.metric_slugs_by_category.return_value = {
                "Sample Category": ['test-metric-a', 'test-metric-b'],
            }
            r.gauge_slugs.return_value = set(['test-gauge'])

            # Do the Request and test for content
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('Sample Category', resp.context['metrics'].keys())

            metrics_values = list(resp.context['metrics'].values())
            self.assertIn('test-metric-a', metrics_values[0])
            self.assertIn('test-metric-b', metrics_values[0])

            # Make sure our Mock R object called the right methods.
            mock_r.assert_has_calls([
                call().metric_slugs_by_category(),
            ])

    def test_metric_detail(self):
        with patch('redis_metrics.views.get_r') as mock_get_r:
            inst = mock_get_r.return_value
            inst._granularities.return_value = ['daily', 'weekly']

            slug = 'test-metric'
            url = reverse('redis_metric_detail', args=[slug])

            # Do the Request & test results
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertEqual(resp.context_data['slug'], slug)
            self.assertEqual(resp.context_data['granularities'], ['daily', 'weekly'])

    def test_metric_history(self):
        with patch('redis_metrics.views.get_r') as mock_get_r:
            inst = mock_get_r.return_value
            inst._granularities.return_value = ['daily', 'weekly']

            slug = 'test-metric'
            granularity = 'daily'
            url = reverse('redis_metric_history', args=[slug, granularity])

            # Do the Request & test results
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            context = resp.context_data
            self.assertEqual(context['slug'], slug)
            self.assertEqual(context['since'], None)
            self.assertEqual(context['granularity'], granularity)
            self.assertEqual(context['granularities'], ['daily', 'weekly'])

    def test_metric_history_since(self):
        """Tests the ``MetricHistoryView`` when there's a ``since`` variable"""
        slug = 'test-metric'
        granularity = 'daily'
        url = reverse('redis_metric_history', args=[slug, granularity])

        # Do the Request & test results for a Date
        resp = self.client.get(url, {'since': '2012-12-25'})
        self.assertEqual(resp.status_code, 200)
        context = resp.context_data
        self.assertEqual(context['since'], datetime(2012, 12, 25))
        self.assertEqual(context['slug'], slug)
        self.assertEqual(context['granularity'], granularity)

        # Do the Request & test results for a Date & Time
        resp = self.client.get(url, {'since': '2012-12-25 11:30:45'})
        self.assertEqual(resp.status_code, 200)
        context = resp.context_data
        self.assertEqual(context['since'], datetime(2012, 12, 25, 11, 30, 45))
        self.assertEqual(context['slug'], slug)
        self.assertEqual(context['granularity'], granularity)

    def test_metrics_list_requires_admin(self):
        """Verifies that ``MetricsListView`` requires authentication."""
        self.assertUnauthedRequestRedirects(reverse('redis_metrics_list'))

    def test_metric_detail_requires_admin(self):
        """Verifies that ``MetricDetailView`` requires authentication."""
        self.assertUnauthedRequestRedirects(
            reverse('redis_metric_detail', args=['whatever'])
        )

    def test_metric_history_requires_admin(self):
        """Verifies that ``MetricHistoryView`` requires authentication."""
        self.assertUnauthedRequestRedirects(
            reverse('redis_metric_history', args=['whatever', 'daily'])
        )

    def test_aggregate_form_view(self):
        """Verifies that GET requests to the ``AggregateFormView`` have the
        correct context info (i.e. a form)."""
        url = reverse('redis_metric_aggregate')
        self.assertUnauthedRequestRedirects(url)
        with patch('redis_metrics.views.get_r'):
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('form', resp.context_data)

    def test_aggregate_form_view_post(self):
        """Verifies that POST requests to the ``AggregateFormView`` work as
        expected."""
        url = reverse('redis_metric_aggregate')
        with patch('redis_metrics.views.get_r'):
            with patch('redis_metrics.forms.R') as mock_r:
                # Set up a return value for ``R.metric_slugs``, which is used
                # in the ``AggregateMetricForm``
                r = mock_r.return_value  # An instance of our Mocked R class
                r.metric_slugs.return_value = set(['test', 'foo', 'bar'])

                # Test with ONE metric selected
                data = {'metrics': ['foo']}
                resp = self.client.post(url, data)
                self.assertEqual(resp.status_code, 302)
                # make sure the metric slug shows up in the redirect URL
                self.assertIn('foo', resp.get("Location", ''))

                # Test with TWO metrics selected
                data = {'metrics': ['foo', 'bar']}
                resp = self.client.post(url, data)
                self.assertEqual(resp.status_code, 302)
                # make sure the metric slug shows up in the redirect URL
                self.assertIn('foo+bar', resp.get("Location", ''))

    def test_aggregate_detail_view(self):
        """Tests ``views.AggregateDetailView``."""
        with patch('redis_metrics.views.get_r') as mock_get_r:
            inst = mock_get_r.return_value
            inst._granularities.return_value = ['daily', 'weekly']

            slug_set = set(['foo', 'bar', 'test-metric', 'yippitty-poo-bah'])
            slugs = '+'.join(slug_set)
            url = reverse('redis_metric_aggregate_detail', args=[slugs])

            # Do the Request & test results
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('slugs', resp.context_data)
            self.assertIn('granularities', resp.context_data)
            self.assertEqual(resp.context_data['slugs'], slug_set)
            self.assertEqual(resp.context_data['granularities'], ['daily', 'weekly'])

    def _test_aggregate_history_view(self, slugs, granularity):
        """Tests ``views.AggregateHistoryView`` with the given slugs,
        granularity, without a specified date verifying the values are
        correctly passed to the template."""
        with patch('redis_metrics.views.get_r') as mock_get_r:
            inst = mock_get_r.return_value
            inst._granularities.return_value = ['daily', 'weekly']

            slug_set = set(slugs)
            url = reverse('redis_metric_aggregate_history',
                          args=['+'.join(slugs), granularity])

            # Do the Request & test results
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('slugs', resp.context_data)
            self.assertIn('granularity', resp.context_data)
            self.assertIn('since', resp.context_data)
            self.assertIn('granularities', resp.context_data)
            self.assertEqual(resp.context_data['slugs'], slug_set)
            self.assertEqual(resp.context_data['granularity'], granularity)
            self.assertEqual(resp.context_data['granularities'], ['daily', 'weekly'])
            self.assertIsNone(resp.context_data['since'])

    def test_aggregate_history_view_hourly(self):
        self._test_aggregate_history_view(['foo', 'bar'], 'hourly')

    def test_aggregate_history_view_daily(self):
        self._test_aggregate_history_view(['foo', 'bar'], 'daily')

    def test_aggregate_history_view_weekly(self):
        self._test_aggregate_history_view(['foo', 'bar'], 'weekly')

    def test_aggregate_history_view_monthly(self):
        self._test_aggregate_history_view(['foo', 'bar'], 'monthly')

    def test_aggregate_history_view_yearly(self):
        self._test_aggregate_history_view(['foo', 'bar'], 'yearly')

    def test_aggregate_history_view_since(self):
        """Tests ``views.AggregateHistoryView`` with a ``since`` parameter."""
        slugs = ['foo', 'bar']
        granularity = 'yearly'
        url = reverse(
            'redis_metric_aggregate_history',
            args=['+'.join(slugs), granularity]
        )

        # Do the Request & test results with a just a date
        resp = self.client.get(url, {'since': "2012-12-25"})
        self.assertEqual(resp.status_code, 200)
        self.assertIsInstance(resp.context_data['since'], datetime)
        self.assertEqual(resp.context_data['since'], datetime(2012, 12, 25))

        # Do the Request & test results with a date & time
        resp = self.client.get(url, {'since': "2012-12-25 11:30:45"})
        self.assertEqual(resp.status_code, 200)
        self.assertIsInstance(resp.context_data['since'], datetime)
        self.assertEqual(
            resp.context_data['since'],
            datetime(2012, 12, 25, 11, 30, 45)
        )

    def test_category_form_view(self):
        """Verifies that GET requests to the ``CategoryFormView`` have the
        correct context info (i.e. a form)."""
        url = reverse('redis_metrics_categorize')
        self.assertUnauthedRequestRedirects(url)

        # NOTE: you can't mock the form in the views, because calls to the
        # form get dispatched somewhere else. That's why this is mocking the
        # R object in the form, instead of the form, itself.
        k = {
            'return_value.metric_slugs.return_value': ['foo', 'bar', 'baz'],
            'return_value._category_slugs.return_value': [],
        }
        with patch('redis_metrics.forms.R', **k):
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('form', resp.context_data)
            self.assertIn("id_category_name", str(resp.content))
            self.assertIn("id_metrics", str(resp.content))

    def test_category_form_view_with_initial(self):
        """Verifies that GET requests to the ``CategoryFormView`` have the
        correct context info/initial data when called with a specified
        Category."""
        url = reverse('redis_metrics_categorize', args=['Stuff'])
        self.assertUnauthedRequestRedirects(url)

        # NOTE: you can't mock the form in the views, because calls to the
        # form get dispatched somewhere else. That's why this is mocking the
        # R object in the form, instead of the form, itself.
        k = {
            'return_value.metric_slugs.return_value': ['foo', 'bar', 'baz'],
            'return_value._category_slugs.return_value': ['foo', 'bar'],
        }
        with patch('redis_metrics.forms.R', **k) as mock_R:
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertIn('form', resp.context_data)
            self.assertIn("id_category_name", str(resp.content))
            self.assertIn("id_metrics", str(resp.content))

            # foo and bar should be pre-selected, but baz should not, since
            # it's not in the "Stuff" category
            initial_metrics = resp.context['form'].fields['metrics'].initial
            self.assertIn('foo', initial_metrics)
            self.assertIn('bar', initial_metrics)
            self.assertNotIn('baz', initial_metrics)

            # This is what should happen in the form
            mock_R.assert_has_calls([
                # happens in __init__
                call(),
                call().metric_slugs(),
                call()._category_slugs('Stuff')
            ])

    def test_category_form_view_post(self):
        """Verifies that POST requests to the ``CategoryFormView`` work as
        expected."""
        url = reverse('redis_metrics_categorize')

        # NOTE: you can't mock the form in the views, because calls to the
        # form get dispatched somewhere else. That's why this is mocking the
        # R object in the form, instead of the form, itself.
        k = {
            'return_value.metric_slugs.return_value': ['foo', 'bar', 'baz'],
            'return_value._category_slugs.return_value': ['foo', 'bar'],
        }
        with patch('redis_metrics.forms.R', **k) as mock_R:
            data = {'category_name': 'Foo', 'metrics': ['foo', 'bar']}
            resp = self.client.post(url, data)
            self.assertEqual(resp.status_code, 302)
            # This is what should happen in the form when POSTing
            mock_R.assert_has_calls([
                # happens in __init__
                call(),
                call().metric_slugs(),

                # happens in categorize_metrics
                call().reset_category('Foo', ['foo', 'bar']),
            ])
